import json
import logging
import random
import subprocess
import time
from typing import Dict
from typing import List
from typing import Optional

import more_itertools

try:
    from winrm import Session
    from winrm.exceptions import WinRMTransportError
    from winrm.vendor.requests_kerberos.exceptions import KerberosExchangeError
except ImportError:
    pass

from .ad_exceptions import CommandFailure
from .ad_exceptions import CprNotFoundInADException
from .ad_exceptions import CprNotNotUnique
from .read_ad_conf_settings import read_settings

logger = logging.getLogger("AdCommon")


ADUser = Dict[str, str]


def ad_minify(text):
    # This function is only used by `ADWriter.remove_redundant`
    text = text.replace("\n", "")
    text = text.replace("\r", "")
    while text.find("  ") > -1:
        text = text.replace("  ", " ")
    return text


def generate_ntlm_session(hostname, system_user, password):
    """Method to create a ntlm session for running powershell scripts.

    Returns:
        winrm.Session
    """
    session = Session(
        "https://{}:5986/wsman".format(hostname),
        transport="ntlm",
        auth=(system_user, password),
        server_cert_validation="ignore",
    )
    return session


class ReauthenticatingKerberosSession:
    """
    Wrapper around WinRM Session object that automatically tries to generate
    a new Kerberos token and session if authentication fails
    """

    def _generate_kerberos_ticket(self):
        """
        Tries to generate a new Kerberos ticket, through a call to kinit
        Raises an exception if the subprocess has non-zero exit code
        """
        cmd = ["kinit", self._username]
        try:
            subprocess.run(
                cmd,
                check=True,
                input=self._password.encode(),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as e:
            print(e.stderr.decode("utf-8"))
            raise

    def _create_new_session(self):
        """
        Generate new internal session

        XXX: Session is not properly cleaned up when a command fails
             so we have to create a new internal session object each time
        """
        self._session = Session(
            target=self._target, transport="kerberos", auth=(None, None)
        )

    def __init__(self, target: str, username: str, password: str):
        self._username = username
        self._password = password
        self._target = target
        self._create_new_session()

    def run_cmd(self, *args, **kwargs):
        try:
            rs = self._session.run_cmd(*args, **kwargs)
        except KerberosExchangeError:
            self._generate_kerberos_ticket()
            self._create_new_session()
            rs = self._session.run_cmd(*args, **kwargs)
        return rs

    def run_ps(self, *args, **kwargs):
        try:
            rs = self._session.run_ps(*args, **kwargs)
        except KerberosExchangeError:
            self._generate_kerberos_ticket()
            self._create_new_session()
            rs = self._session.run_ps(*args, **kwargs)
        return rs


def generate_kerberos_session(hostname, username=None, password=None):
    """
    Method to create a kerberos session for running powershell scripts.

    Returned object will have same public interface as WinRM Session, despite
    not inheriting from it

    Returns:
        ReauthenticatingKerberosSession
    """
    session = ReauthenticatingKerberosSession(
        "http://{}:5985/wsman".format(hostname),
        username=username,
        password=password,
    )
    return session


class AD:
    _encoding = "utf-8"

    def __init__(self, all_settings=None, index=0, **kwargs):
        self.all_settings = all_settings
        if self.all_settings is None:
            self.all_settings = read_settings(index=index)
        self.session = self._create_session()
        self.retry_exceptions = self._get_retry_exceptions()
        self.results = {}

    def _get_retry_exceptions(self):
        """Tuple of exceptions which should trigger retrying create_session."""
        return (WinRMTransportError,)

    def _create_session(self):
        """Method to create a session for running powershell scripts.

        The returned object should have a run_ps method, which consumes a
        powershell script, and returns a status object.

        The status object should have a status_code, std_out and std_err
        attribute, containing the result from executing the powershell script.

        Returns:
            winrm.Session
        """

        all_settings = self.all_settings

        if all_settings["primary"]["method"] == "ntlm":
            session = generate_ntlm_session(
                all_settings["global"]["winrm_host"],
                all_settings["global"]["system_user"],
                all_settings["global"]["password"],
            )
        elif all_settings["primary"]["method"] == "kerberos":
            session = generate_kerberos_session(
                all_settings["global"]["winrm_host"],
                all_settings["global"]["system_user"],
                all_settings["global"]["password"],
            )
        else:
            raise ValueError(
                "Unknown WinRM method: %r" % all_settings["primary"]["method"]
            )

        return session

    def _run_ps_script(self, ps_script):
        """Run a PowerShell script and return the result.

        If the script fails, a `CommandFailure` exception is raised.
        If the script output cannot be parsed as JSON, a `ValueError` exception
        is raised.

        :param ps_script: The PowerShell script to run.
        :return: A dictionary with the returned parameters.
        """

        encoding = self._ps_boiler_plate()["encoding"]
        if encoding not in ps_script:
            sep = "\n" if not ps_script.startswith("\n") else ""
            ps_script = f"{encoding}{sep}{ps_script}"

        logger.debug("Attempting to run script: {}".format(ps_script))
        response = {}
        if not self.session:
            return response

        retries = 0
        try_again = True
        while try_again and retries < 10:
            try:
                r = self.session.run_ps(ps_script)
                try_again = False
            except self.retry_exceptions:
                logger.error("AD read error: {}".format(retries))
                time.sleep(5)
                retries += 1
                # The existing session is now dead, create a new.
                self.session = self._create_session()

        # TODO: We will need better error handling than this.
        assert retries < 10

        if r.status_code != 0:
            raise CommandFailure(r.std_err)
        if not r.std_out:
            logger.warning("status_code=0 but no std_out")
            return {}
        return self._parse_ps_script_result(ps_script, r)

    def _parse_ps_script_result(self, script, response):
        output = response.std_out
        if isinstance(output, bytes):
            output = output.decode(self._encoding)

        try:
            return json.loads(output)
        except json.JSONDecodeError as exc:
            msg = "Could not parse JSON response, error in lines:\n%s\nscript:\n%s"
            line_sep = b"\n" if isinstance(response.std_out, bytes) else "\n"
            invalid_bound = slice(max(0, exc.lineno - 1), exc.lineno + 1)
            invalid_lines = response.std_out.splitlines()[invalid_bound]
            invalid_lines = line_sep.join(invalid_lines)
            logger.error(msg, invalid_lines, script)
            raise ValueError(msg % (invalid_lines, script))

    def _ps_boiler_plate(self):
        """
        Boiler plate that needs to go into all PowerShell code.
        """
        settings = self._get_setting()

        # TODO: Are these really different?
        path = ""
        search_base = ""
        if settings["search_base"]:
            path = ' -Path "{}" '.format(settings["search_base"])
            search_base = ' -SearchBase "{}" '.format(settings["search_base"])

        credentials = " -Credential $usercredential"

        boiler_plate = {
            "encoding": "[Console]::OutputEncoding = [Text.UTF8Encoding]::UTF8",
            "path": path,
            "search_base": search_base,
            "credentials": credentials,
            "complete": search_base + credentials,
        }
        return boiler_plate

    def remove_redundant(self, text):
        # This method is poorly named, as is `ad_minify`. `ad_minify` actually removes
        # linebreaks and superfluous whitespace from `text`, which is assumed to be a
        # PowerShell command fragment.
        return ad_minify(text)

    def _build_ps(self, ps_script, format_rules):
        """
        Return the standard code need to execute a power shell script from a
        template.
        """
        # This method is only used by `ADWriter._get_add_manager_command`.
        # TODO: Consider removing this method and integrating its logic in
        # `ADWriter._get_add_manager_command` instead.
        formatted_script = ps_script.format(**format_rules)
        finished_ps_script = self._build_user_credential() + self.remove_redundant(
            formatted_script
        )
        return finished_ps_script

    def _get_setting(self):
        return self.all_settings["primary"]

    def _build_user_credential(self):
        """
        Build the commonn set of Power Shell commands that is needed to
        run the AD commands.
        :return: A suitable string to prepend to AD commands.
        """

        credential_template = """
        $User = "{}"
        $PWord = ConvertTo-SecureString –String "{}" –AsPlainText -Force
        $TypeName = "System.Management.Automation.PSCredential"
        $UserCredential = New-Object –TypeName $TypeName –ArgumentList $User, $PWord
        """
        settings = self._get_setting()
        user_credential = credential_template.format(
            settings["system_user"], settings["password"]
        )
        return user_credential

    def _unreadable_properties(self) -> List[str]:
        """Return a list of AD properties which cannot be read by `Get-ADUser`, as they
        are write-only."""
        return ["AccountPassword"]

    def _properties(self):
        # This is only called from `AD.get_from_ad`, and as such, it is only used to
        # create a list of the AD properties we want to *read* (not write.)
        # Thus it is irrelevant for the code paths which create or update AD users.
        settings = self._get_setting()
        properties = " -Properties "
        for item in settings["properties"]:
            # Skip unreadable AD properties, such as "AccountPassword"
            if item in self._unreadable_properties():
                continue
            properties += item + ","
        properties = properties[:-1] + " "  # Remove trailing comma, add space
        return properties

    def _get_sam_from_ad_values(self, ad_values: ADUser) -> str:
        # `ad_values` contains a dict with data on *one* AD user
        return ad_values["SamAccountName"]

    def _find_ad_user(
        self, cpr: str, ad_dump: Optional[List[Dict[str, str]]] = None
    ) -> ADUser:
        """Find a unique AD account from cpr, otherwise raise an exception."""
        if ad_dump:
            cpr_field = self.all_settings["primary"]["cpr_field"]
            ad_users = filter(lambda ad_user: ad_user.get(cpr_field) == cpr, ad_dump)
        else:
            logger.debug("No AD information supplied, will look it up")
            ad_users = self.get_from_ad(cpr=cpr)

        return more_itertools.one(ad_users, CprNotFoundInADException, CprNotNotUnique)

    def get_from_ad(self, user=None, cpr=None, server=None):
        """
        Read all properties of an AD user. The user can be retrived either by cpr
        or by AD user name.

        Example:

            [
                {
                    'ObjectGUID': '7ccbd9aa-gd60-4fa1-4571-0e6f41f6ebc0',
                    'SID': {
                        'AccountDomainSid': {
                            'AccountDomainSid': 'S-x-x-xx-...',
                            'BinaryLength': 24,
                            'Value': 'S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx'
                        },
                        'BinaryLength': 28,
                        'Value': 'S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx-xxxxx'
                    },
                    'PropertyCount': 11,
                    'PropertyNames': [
                        'ObjectGUID',
                        'SID',
                        'DistinguishedName',
                        'Enabled',
                        'GivenName',
                        'Name',
                        'ObjectClass',
                        'SamAccountName',
                        'Surname',
                        'UserPrincipalName'
                        'extensionAttribute1',
                    ],
                    'DistinguishedName': 'CN=Martin Lee Gore,OU=...'
                    'Enabled': True,
                    'GivenName': 'Martin Lee',
                    'Name': 'Martin Lee Gore',
                    'ObjectClass': 'user',
                    'SamAccountName': 'mlego',
                    'Surname': 'Gore',
                    'UserPrincipalName': 'martinleegore@magenta.dk',
                    'extensionAttribute1': '1122334455',
                    'AddedProperties': [],
                    'ModifiedProperties': [],
                    'RemovedProperties': [],
                }
            ]

        :param user: The SamAccountName to retrive.
        :param cpr: cpr number of the user to retrive.
        :param server: Add an explcit server to the query. Mostly needed to check
        if replication is finished.
        :return: All properties listed in AD for the user.
        """
        settings = self._get_setting()
        bp = self._ps_boiler_plate()

        if user:
            ps_template = "Get-ADUser -Filter 'SamAccountName -eq \"{val}\"'"
            get_command = ps_template.format(val=user)

        if cpr:
            # For wild-card searches, we do a litteral search.
            if cpr.find("*") > -1:
                val = cpr
            else:
                # For direct cpr-search we obey the local separator setting.
                cpr_sep = settings["cpr_separator"]
                val = f"{cpr[0:6]}{cpr_sep}{cpr[6:10]}"

            field = settings["cpr_field"]
            operator = "-eq" if field.lower() == "objectguid" else "-like"
            ps_template = "Get-ADUser -Filter '{field} {operator} \"{val}\"'"
            get_command = ps_template.format(field=field, operator=operator, val=val)

        server_string = ""
        if server is not None:
            server_string = " -Server {}".format(server)
        elif self.all_settings["global"].get("servers"):
            server_string = " -Server {}".format(
                random.choice(self.all_settings["global"]["servers"])
            )

        ps_script = (
            self._ps_boiler_plate()["encoding"]
            + self._build_user_credential()
            + get_command
            + server_string
            + bp["complete"]
            + self._properties()
            + " | ConvertTo-Json"
        )

        response = self._run_ps_script(ps_script)

        if not response:
            return_val = []
        else:
            if not isinstance(response, list):
                return_val = [response]
            else:
                return_val = response

        return return_val



# SPDX-FileCopyrightText: Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
from functools import lru_cache
from typing import Any

from pydantic import AnyHttpUrl
from pydantic import BaseSettings
from pydantic import Field
from pydantic import SecretStr
from ra_utils.job_settings import JobSettings


class GqlLoraCacheSettings(JobSettings):  # type: ignore
    class Config:
        frozen = True

    use_new_cache: bool = False
    primary_manager_responsibility: str | None = None
    prometheus_pushgateway: str = "pushgateway"
    mox_base: str = "http://mo:5000/lora"
    std_page_size: int = 300

    def to_old_settings(self) -> dict[str, Any]:
        """Convert our DatabaseSettings to a settings.json format.

        This serves to implement the adapter pattern, adapting from pydantic and its
        corresponding 12-factor configuration paradigm with environment variables, to
        the current functionality of the program, based on the settings format from
        settings.json.

        Eventually the entire settings-processing within the program should be
        rewritten with a process similar to what has been done for the SD integration,
        but it was out of scope for the change when this code was introduced.
        """

        settings = {
            "mora.base": self.mora_base,
            "mox.base": self.mox_base,
            "exporters": {
                "actual_state": {
                    "manager_responsibility_class": self.primary_manager_responsibility
                }
            },
            "exporters.actual_state.manager_responsibility_class": self.primary_manager_responsibility,
            "use_new_cache": self.use_new_cache,
        }

        return settings


@lru_cache()
def get_gql_cache_settings(*args, **kwargs) -> GqlLoraCacheSettings:
    return GqlLoraCacheSettings(*args, **kwargs)


class DatabaseConfiguration(BaseSettings):  # type: ignore
    class Config:
        frozen = True

    type: str | None
    host: str | None
    db_name: str
    user: str | None
    password: SecretStr | None


class DatabaseSettings(JobSettings):
    class Config:
        frozen = True
        env_nested_delimiter = "__"

    mora_base: AnyHttpUrl = Field("http://mo-service:5000")
    mox_base: AnyHttpUrl = Field("http://mo-service:5000/lora")

    actual_state: DatabaseConfiguration = Field(default_factory=DatabaseConfiguration)
    historic_state: DatabaseConfiguration | None
    log_overlapping_aak: bool = False
    use_new_cache: bool = False
    primary_manager_responsibility: str | None = None

    def to_old_settings(self) -> dict[str, Any]:
        """Convert our DatabaseSettings to a settings.json format.

        This serves to implement the adapter pattern, adapting from pydantic and its
        corresponding 12-factor configuration paradigm with environmental variables, to
        the current functionality of the program, based on the settings format from
        settings.json.

        Eventually the entire settings-processing within the program should be
        rewritten with a process similar to what has been done for the SD integration,
        but it was out of scope for the change when this code was introduces.
        """

        def secret_val_or_none(value: SecretStr | None) -> str | None:
            if value is None:
                return None
            return value.get_secret_value()

        settings = {
            "mora.base": self.mora_base,
            "mox.base": self.mox_base,
            "exporters.actual_state.type": self.actual_state.type,
            "exporters.actual_state.host": self.actual_state.host,
            "exporters.actual_state.db_name": self.actual_state.db_name,
            "exporters.actual_state.user": self.actual_state.user,
            "exporters.actual_state.password": secret_val_or_none(
                self.actual_state.password
            ),
            "primary_manager_responsibility": self.primary_manager_responsibility,
            "exporters.actual_state.manager_responsibility_class": self.primary_manager_responsibility,
            "use_new_cache": self.use_new_cache,
        }
        if self.historic_state is not None:
            settings.update(
                {
                    "exporters.actual_state_historic.type": self.historic_state.type
                    or self.actual_state.type,
                    "exporters.actual_state_historic.host": self.historic_state.host
                    or self.actual_state.host,
                    "exporters.actual_state_historic.db_name": self.historic_state.db_name
                    or self.actual_state.db_name,
                    "exporters.actual_state_historic.user": self.historic_state.user
                    or self.actual_state.user,
                    "exporters.actual_state_historic.password": secret_val_or_none(
                        self.historic_state.password
                    )
                    or secret_val_or_none(self.actual_state.password),
                    "primary_manager_responsibility": self.primary_manager_responsibility,
                    "use_new_cache": self.use_new_cache,
                }
            )
        return settings



import copy
import inspect
import logging
import re
from datetime import datetime
from unittest import mock
from unittest import TestCase
from unittest.mock import MagicMock

from freezegun import freeze_time
from hypothesis import given
from hypothesis import strategies as st
from jinja2.exceptions import UndefinedError
from more_itertools import first_true
from more_itertools import only
from os2mo_helpers.mora_helpers import MoraHelper
from parameterized import parameterized
from ra_utils.lazy_dict import LazyDict

from ..ad_exceptions import CommandFailure
from ..ad_exceptions import CprNotFoundInADException
from ..ad_exceptions import CprNotNotUnique
from ..ad_exceptions import NoPrimaryEngagementException
from ..ad_exceptions import SamAccountNameNotUnique
from ..ad_template_engine import illegal_parameters
from ..ad_template_engine import INVALID
from ..ad_writer import ADWriter
from ..ad_writer import LoraCacheSource
from ..user_names import UserNameSetInAD
from ..utils import AttrDict
from .mocks import MO_MANAGER_CPR
from .mocks import MO_MANAGER_SAM
from .mocks import MO_MANAGER_UUID
from .mocks import MO_ROOT_ORG_UNIT_NAME
from .mocks import MO_USER_CPR
from .mocks import MO_USER_SAM
from .mocks import MO_UUID
from .mocks import MockADParameterReaderWithManager
from .mocks import MockADWriterContext
from .mocks import MockLoraCacheUnitAddress
from .mocks import MockLoraCacheWithManager
from .mocks import MockMOGraphqlSource
from .mocks import MockMORESTSource
from .test_utils import dict_modifier
from .test_utils import mo_modifier
from .test_utils import TestADWriterMixin


JOB_TITLE_AD_FIELD_NAME = "titel"
JOB_TITLE_TEMPLATE = "{{ ad_values.get('titel') or mo_values['title'] }}"

_SYNC_TIMESTAMP = "2020-01-01 12:30:00"


class TestADWriter(TestCase, TestADWriterMixin):
    def setUp(self):
        self._setup_adwriter()

    def _verify_identical_common_code(self, num_expected_scripts, num_common_lines=5):
        """Verify that common code in all scripts is identical.

        I.e. that all scripts start with the same num_common_lines lines.
        """
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)
        # 1. Convert each script from a string into a list of strings (lines)
        lines = [script.split("\n") for script in self.ad_writer.scripts]
        self.assertGreaterEqual(len(lines[0]), num_common_lines)
        self.assertEqual(len(lines), num_expected_scripts)
        # 2. Get the common lines (first `num_common_lines` lines in each
        # script.)
        common_lines = [x[:num_common_lines] for x in lines]
        self.assertEqual(len(common_lines[0]), num_common_lines)
        self.assertEqual(len(common_lines), num_expected_scripts)
        # 3. Zip the lines producing `num_common_lines` tuples of 'n' elements,
        # where 'n' is len(scripts).
        zip_lines = list(zip(*common_lines))
        self.assertEqual(len(zip_lines[0]), num_expected_scripts)
        self.assertEqual(len(zip_lines), num_common_lines)
        # Check that all zip_lines are identical
        for zip_line in zip_lines:
            self.assertEqual(len(set(zip_line)), 1)
        # Return common code
        return common_lines[0]

    def _add_to_template_to_ad_fields(self, ad_field_name, template):
        return dict_modifier(
            {"integrations.ad_writer.template_to_ad_fields": {ad_field_name: template}}
        )

    def _assert_script_contains_field(self, script, name, value):
        """Assert that AD PowerShell script `script` includes the AD field
        called `name` and that the field value is `value`.
        """
        self.assertRegex(script, f'"{name}"="{value}"')

    def _get_script_contents(self, index=0):
        return self.ad_writer.scripts[index].split("\n")[5].strip()

    @parameterized.expand(
        [
            # Test without any changes to settings
            [
                dict_modifier({}),
            ],
            # Test with overridden password
            [
                dict_modifier({"primary": {"password": "Password1"}}),
            ],
            [
                dict_modifier({"primary": {"password": "Hunter2"}}),
            ],
            # Test with overridden user
            [
                dict_modifier({"primary": {"system_user": "R2D2"}}),
            ],
            [
                dict_modifier({"primary": {"system_user": "C-3PO"}}),
            ],
        ]
    )
    def test_common_ps_code(self, settings_transformer):
        """Test ps_script common code (first five lines of each script).

        The common code lines are identical for all writes to the AD.
        This is verified by the 'verify_identical_common_code' method.

        Create_user is used as a test-case to provoke the creation of a PS script,
        but only the common top of the code is actually tested.
        """
        self._setup_adwriter(settings_transformer)

        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.create_user(mo_uuid=uuid, create_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        common_ps = self._verify_identical_common_code(num_expected_scripts)
        common_ps = [x.strip() for x in common_ps]
        expected_ps = [
            self.ad_writer._ps_boiler_plate()["encoding"],
            '$User = "' + self.settings["primary"]["system_user"] + '"',
            '$PWord = ConvertTo-SecureString –String "'
            + self.settings["primary"]["password"]
            + '" –AsPlainText -Force',
            '$TypeName = "System.Management.Automation.PSCredential"',
            (
                "$UserCredential = New-Object –TypeName $TypeName "
                "–ArgumentList $User, $PWord"
            ),
        ]
        self.assertEqual(common_ps, expected_ps)

    # Jinja to determine if mo_values['employment_number'] is a prime number
    is_prime_jinja = """
        {#- Aliased to n for readability -#}
        {%- set n = mo_values['employment_number'] | int -%}
        {#- Negative numbers, zero and one are not primes -#}
        {%- if n <= 1 -%}
            false
        {%- else -%}
            {#- Using vars as dict allows settings by reference -#}
            {%- set vars = {'is_prime': True} -%}

            {#- Really we only need to check until sqrt(n) -#}
            {%- for i in range(2,n) -%}
                {#- We cannot use percentage for modulus -#}
                {%- set quotient = (n/i) | int -%}
                {%- set remainder = n - i*quotient -%}

                {#- If it divides without remainder, it is not prime -#}
                {%- if remainder == 0 -%}
                    {#- Update is_prime by reference -#}
                    {%- set _ = vars.update({'is_prime': False}) -%}
                {%- endif -%}
            {%- endfor -%}

            {#- Print whether we found a prime or not -#}
            {%- if vars['is_prime'] -%}
                true
            {%- else -%}
                false
            {%- endif -%}
        {%- endif -%}
    """

    @parameterized.expand(
        [
            # Test without any changes
            [dict_modifier({}), None, ""],
            # Test with new employment number
            [dict_modifier({}), mo_modifier({"employment_number": "42"}), ""],
            [dict_modifier({}), mo_modifier({"employment_number": "100"}), ""],
            # Test mo_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.mo_to_ad_fields": {
                            "unit_user_key": "unit_user_key"
                        }
                    }
                ),
                None,
                '"unit_user_key"="Musik";',
            ],
            # Test with added template_field
            # Simple field lookup
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "unit_user_key": "{{ mo_values['unit_user_key'] }}"
                        }
                    }
                ),
                None,
                '"unit_user_key"="Musik";',
            ],
            # Test with timestamp as added template_field
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "extensionAttribute21": "{{ sync_timestamp }}",
                        }
                    }
                ),
                None,
                f'"extensionAttribute21"="{_SYNC_TIMESTAMP}";',
            ],
            # Field lookup and processing
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "street_number": """
                                {%- set street = mo_values['unit_streetname'] -%}
                                {{ street.split(' ')[-1] }}
                                {#- Comment used to trim whitespace -#}
                            """
                        }
                    }
                ),
                None,
                '"street_number"="451";',
            ],
            # Test if employment number is prime
            # 101 is prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "101"}),
                '"employment_number_is_prime"="true";',
            ],
            # 100 is not prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "100"}),
                '"employment_number_is_prime"="false";',
            ],
            # 263 is prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "263"}),
                '"employment_number_is_prime"="true";',
            ],
            # 267 is not prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "267"}),
                '"employment_number_is_prime"="false";',
            ],
        ]
    )
    @freeze_time(_SYNC_TIMESTAMP)
    def test_user_create_custom_fields(
        self, settings_transformer, mo_transformer, expected
    ):
        """Test user create ps_script code.

        The common code is not tested.
        """
        self._setup_adwriter(None, mo_transformer, settings_transformer)
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.create_user(mo_uuid=uuid, create_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        create_user_ps = self._get_script_contents(index=0)

        mo_values = self.ad_writer.read_ad_information_from_mo(uuid)

        if "extensionAttribute21" in expected:
            # When checking datetime we need to drop the seconds before comparing
            expected = expected[:-11]

        expected_content = [
            "New-ADUser",
            '-Name "'
            + mo_values["full_name"]
            + " - "
            + mo_values["sam_account_name"]
            + '"',
            '-Displayname "' + mo_values["full_name"] + '"',
            '-GivenName "' + mo_values["name"][0] + '"',
            '-SurName "' + mo_values["name"][1] + '"',
            '-SamAccountName "' + mo_values["sam_account_name"] + '"',
            '-EmployeeNumber "' + mo_values["employment_number"] + '"',
            "-Credential $usercredential",
            '-UserPrincipalName "' + mo_values["sam_account_name"] + '@epn_end"',
            "-OtherAttributes",
            '"level2orgunit_field"="Ingen";',
            '"org_field"="Kommune\\Forvalting\\Enhed\\";',
            '"uuid_field"="' + mo_values["uuid"] + '";',
            '"cpr_field"="'
            + mo_values["cpr"][0:6]
            + "ad_cpr_sep"
            + mo_values["cpr"][6:]
            + '"',
            '-Path "search_base"',
            expected,
        ]
        for content in expected_content:
            self.assertIn(content, create_user_ps)

    def test_user_edit(self):
        """Test user edit ps_script code.

        The common code is not tested.
        """
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.sync_user(mo_uuid=uuid, sync_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        edit_user_ps = self._get_script_contents()
        mo_values = self.ad_writer.read_ad_information_from_mo(uuid)
        expected_content = [
            "Get-ADUser",
            "-Filter 'SamAccountName -eq \"" + mo_values["sam_account_name"] + "\"'",
            "-Credential $usercredential",
            "|",
            "Set-ADUser",
            "-Credential $usercredential",
            '-Displayname "' + mo_values["full_name"] + '"',
            '-GivenName "' + mo_values["name"][0] + '"',
            '-SurName "' + mo_values["name"][1] + '"',
            '-EmployeeNumber "' + mo_values["employment_number"] + '"',
            "-Replace",
            '"level2orgunit_field"="Ingen";',
            '"org_field"="Kommune\\Forvalting\\Enhed\\";',
        ]
        for content in expected_content:
            self.assertIn(content, edit_user_ps)

    def test_user_create_ad_values(self):
        """Test that `ad_values` is present (and empty) when creating a new AD
        user. The template `JOB_TITLE_TEMPLATE` uses the MO job title to fill
        in the job title in the appropriate AD field.
        """
        self._setup_adwriter(
            early_transform_settings=self._add_to_template_to_ad_fields(
                JOB_TITLE_AD_FIELD_NAME,
                JOB_TITLE_TEMPLATE,
            )
        )
        mo_uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.create_user(mo_uuid, create_manager=False)
        # Find the "create AD user" script
        create_script = first_true(
            self.ad_writer.scripts, pred=lambda s: "New-ADUser" in s
        )
        # Assert that it contains the mapped job title field
        self._assert_script_contains_field(
            create_script, JOB_TITLE_AD_FIELD_NAME, self.mo_values_func()["title"]
        )

    @parameterized.expand(
        [
            # Tuples of (AD job title, which value to use)
            (None, "mo-value"),
            ("", "mo-value"),
            ("Tester", "ad-value"),
        ]
    )
    def test_user_update_ad_values(self, ad_value, expectation):
        """Test that `ad_values` is present (and empty) when updating an
        existing AD user. The template `JOB_TITLE_TEMPLATE` only populates the
        AD job title with the MO job title if the current AD job title is
        empty.
        """
        self._setup_adwriter(
            early_transform_settings=self._add_to_template_to_ad_fields(
                JOB_TITLE_AD_FIELD_NAME,
                JOB_TITLE_TEMPLATE,
            ),
            transform_ad_values=dict_modifier(
                {JOB_TITLE_AD_FIELD_NAME: ad_value},
            ),
        )
        mo_uuid = "invalid-provided-and-accepted-due-to-mocking"
        mo_value = self.mo_values_func()["title"]
        self.ad_writer.sync_user(mo_uuid, ad_dump=None, sync_manager=False)
        # Find the "update AD user" script
        update_script = first_true(
            self.ad_writer.scripts, pred=lambda s: "Set-ADUser" in s
        )
        # Assert that the update script uses the current AD job title
        self._assert_script_contains_field(
            update_script,
            JOB_TITLE_AD_FIELD_NAME,
            ad_value if ad_value else mo_value,
        )

    def test_duplicate_ad_field_entries(self):
        """Test user edit ps_script code.

        The common code is not tested.
        """
        # These keys conflict, and thus, no primary_write settings are emitted
        settings_transformer = dict_modifier(
            {
                "integrations.ad_writer.template_to_ad_fields": {
                    "Name": "{{ mo_values['unit'] }}"
                },
                "integrations.ad_writer.mo_to_ad_fields": {"unit": "name"},
            }
        )
        with self.assertRaises(ValueError):
            self._setup_adwriter(early_transform_settings=settings_transformer)

    def test_non_overwritten_default(self):
        """Test that a mistake in overriding a default results in an error."""
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Add another mapping for the display name AD field
        # In `TestADMixin._prepare_settings.default_settings`, this is spelled
        # `Displayname` (notice the case difference.)
        display_name = {"displayname": "{{ mo_values['name'][0] }}"}
        self.settings["primary_write"]["template_to_ad_fields"].update(display_name)

        with self.assertRaises(ValueError):
            self.ad_writer.sync_user(
                mo_uuid="invalid-provided-and-accepted-due-to-mocking",
                sync_manager=False,
            )

    def test_user_edit_illegal_parameter(self):
        """Test user edit ps_script code with illegal parameter

        The common code is not tested.

        This test simply ensures that the illegal parameter is dropped.
        """
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)
        illegal_parameters["Set-ADUser"].append("Displayname")

        settings_transformer = dict_modifier({})
        self._setup_adwriter(settings_transformer)
        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.sync_user(mo_uuid=uuid, sync_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        edit_user_ps = self._get_script_contents()
        self.assertNotIn("Displayname", edit_user_ps)

    def test_user_edit_illegal_attribute(self):
        """Test user edit ps_script code with illegal attribute.

        The common code is not tested.

        This test simply ensures that the illegal attribute is dropped.
        """
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        settings_transformer = dict_modifier(
            {"integrations.ad_writer.template_to_ad_fields": {"Name": "John"}}
        )

        self._setup_adwriter(settings_transformer)
        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.sync_user(mo_uuid=uuid, sync_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        edit_user_ps = self._get_script_contents()
        self.assertNotIn('"Name"="John"', edit_user_ps)

    def test_duplicated_ad_field_name(self):
        # Test configuration where 'Name' is mapped both via
        # `mo_to_ad_fields` *and* `template_to_ad_fields`
        add_dupe_field = dict_modifier(
            {
                "integrations.ad_writer.mo_to_ad_fields": {"unit": "Name"},
            }
        )
        with self.assertRaises(ValueError):
            self._setup_adwriter(early_transform_settings=add_dupe_field)

    @parameterized.expand(
        [
            # Verify different employment numbers
            [
                dict_modifier({}),
                mo_modifier({"employment_number": "267"}),
                dict_modifier({}),
                [],
            ],
            [
                dict_modifier({}),
                mo_modifier({"employment_number": "42"}),
                dict_modifier({}),
                [],
            ],
            # Test mo_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.mo_to_ad_fields": {"unit": "Name"},
                    }
                ),
                mo_modifier({"employment_number": "42"}),
                dict_modifier(
                    {
                        "name": ("John Deere", "Enhed"),
                    }
                ),
                ["Name"],
            ],
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.mo_to_ad_fields": {
                            "unit": "Name",
                            "employment_number": "extension_field2",
                        },
                    }
                ),
                mo_modifier({"employment_number": "42"}),
                dict_modifier(
                    {
                        "name": ("John Deere", "Enhed"),
                        "extension_field2": (None, "42"),
                    }
                ),
                ["Name"],
            ],
            # Test template_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "adjusted_number": (
                                "{{ mo_values['employment_number']|int + 5 }}"
                            )
                        },
                    }
                ),
                mo_modifier({"employment_number": "42"}),
                dict_modifier(
                    {
                        "adjusted_number": (None, "47"),
                    }
                ),
                [],
            ],
            # Test template_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "adjusted_number": (
                                "{{ mo_values['employment_number']|int + 5 }}"
                            ),
                            "Enabled": "Invalid",
                        },
                    }
                ),
                mo_modifier(
                    {
                        "employment_number": "42",
                    }
                ),
                dict_modifier(
                    {
                        "adjusted_number": (None, "47"),
                        "enabled": (True, "Invalid"),
                    }
                ),
                [],
            ],
        ]
    )
    def test_sync_compare(
        self,
        settings_transformer,
        mo_transformer,
        expected_transformer,
        removed_fields,
    ):
        # If test case defines `removed_fields`, remove these fields from the
        # `integrations.ad_writer.template_to_ad_fields` setting, as they are
        # defined by the `integrations.ad_writer.mo_to_ad_fields` setting
        # instead.
        def remove_template_fields(settings):
            key = "integrations.ad_writer.template_to_ad_fields"
            to_remove = set([f.lower() for f in removed_fields])
            for field in copy.copy(settings[key]):
                if field.lower() in to_remove:
                    del settings[key][field]
            return settings

        def actual_settings_transformer(settings):
            return settings_transformer(remove_template_fields(settings))

        self._setup_adwriter(
            early_transform_settings=actual_settings_transformer,
            transform_mo_values=mo_transformer,
        )

        uuid = "invalid-provided-and-accepted-due-to-mocking"
        mo_values = self.ad_writer.read_ad_information_from_mo(uuid)
        mo_values["manager_cpr"] = None
        ad_values = self._prepare_get_from_ad(lambda x: x)
        ad_values["Name"] = "John Deere"

        def find_ad_user(cpr, ad_dump=None):
            return ad_values

        self.ad_writer._find_ad_user = find_ad_user

        mismatch = self.ad_writer._sync_compare(mo_values, None)
        expected = {
            "level2orgunit_field": (None, "Ingen"),
            "org_field": (None, "Kommune\\Forvalting\\Enhed\\"),
            "name": (ad_values["Name"], mo_values["name_sam"]),
            "displayname": (None, " ".join(mo_values["name"])),
            "givenname": (ad_values["GivenName"], mo_values["name"][0]),
            "surname": (None, mo_values["name"][1]),
            "employeenumber": (None, mo_values["employment_number"]),
        }
        expected = expected_transformer(expected)
        self.assertEqual(mismatch, expected)

        # Apply changes to AD, and check _sync_compare returns nothing
        ad_keys = ad_values.keys()
        for lower_ad_key, changeset in expected.items():
            cased_ad_key = next(
                filter(lambda ad_key: ad_key.lower() == lower_ad_key, ad_keys),
                lower_ad_key,
            )
            ad_values[cased_ad_key] = changeset[1]
        mismatch = self.ad_writer._sync_compare(mo_values, None)
        self.assertEqual(mismatch, {})

    def test_sync_compare_context_includes_ad_values(self):
        self._setup_adwriter()

        # The `ad_values` passed in the template context has all keys in lower
        # case
        ad_values = {key.lower(): val for key, val in self.ad_values_func().items()}

        mo_values = self.ad_writer.read_ad_information_from_mo("mo-uuid")
        mo_values["manager_cpr"] = None

        # Call `_sync_compare` with a mocked `_render_field_template` method
        # so we can inspect the template contexts passed to it
        with mock.patch.object(self.ad_writer, "_render_field_template") as m:
            self.ad_writer._sync_compare(mo_values, None)

        # Assert that all calls to mocked template render method includes a
        # valid `ad_values` dict in the template context
        template_contexts = [
            call.args[0] for call in m.mock_calls if len(call.args) == 2
        ]
        self.assertTrue(
            all(context["ad_values"] == ad_values for context in template_contexts)
        )

    @mock.patch("time.sleep")
    def test_rename_ad_user(self, *args):
        # Configure ADWriter to write an AD "Name" attribute consisting of the
        # user's full name, plus " testing".
        settings_transformer = dict_modifier(
            {
                "integrations.ad_writer.template_to_ad_fields": {
                    "Name": "{{ mo_values['full_name'] }} - testing"
                },
            }
        )

        # TODO: this duplicates code found in "test_mo_to_ad_sync.py".
        def remove_manager_cpr(mo_values, *args, **kwargs):
            del mo_values["manager_cpr"]
            return mo_values

        self._setup_adwriter(
            early_transform_settings=settings_transformer,
            transform_mo_values=remove_manager_cpr,
        )

        # Construct a mock AD dump, where the user has a Name attr consisting
        # of their full name only.
        ad_dump = [self._prepare_get_from_ad(ad_transformer=None)]

        # Invoke the AD rename
        mo_user_uuid = "not-really-a-uuid"
        self.ad_writer.sync_user(mo_user_uuid, ad_dump=ad_dump)

        # Assert that we issued two AD PowerShell commands: one for the rename,
        # and one for updating the other AD user attributes.
        self.assertEqual(len(self.ad_writer.scripts), 2)

        # Fetch the "rename" command itself: the last line of the first script
        rename_cmd = self.ad_writer.scripts[0].splitlines()[-1]
        # Assert that the "rename" command does indeed attempt to rename the AD
        # user to "Firstname Lastname - testing", due to the mapping in
        # `settings_transformer`.
        expected_new_name = "%s - testing" % ad_dump[0]["Name"]
        expected_cmd_fragment = '-NewName "%s"' % expected_new_name
        self.assertIn(expected_cmd_fragment, rename_cmd)

    def test_add_manager(self):
        mo_values = self.ad_writer.read_ad_information_from_mo(
            uuid="0", read_manager=True
        )
        self.ad_writer.add_manager_to_user(
            "MGORE", manager_sam=mo_values["manager_sam"]
        )
        # Expected outputs
        num_expected_scripts = 1
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)
        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)
        # Check that the create user ps looks good
        add_manager_ps = self._get_script_contents()

        expected_line = (
            "Get-ADUser -Filter 'SamAccountName -eq \"MGORE\"'"
            + " -Credential $usercredential |Set-ADUser -Manager "
            + mo_values["manager_sam"]
            + " -Credential $usercredential"
        )
        self.assertEqual(add_manager_ps, expected_line)

    @parameterized.expand(
        [
            # AD dump is None
            (None, CprNotFoundInADException),
            # AD dump is empty list
            ([], CprNotFoundInADException),
            # AD dump has user without CPR
            ([{"foo": "bar"}], CprNotFoundInADException),
            # AD dump has exactly one matching user
            ([{"cpr": "112233-4455"}], None),
            # AD dump has multiple users with identical CPRs
            (
                [{"cpr": "112233-4455"}, {"cpr": "112233-4455"}],
                CprNotNotUnique,
            ),
        ]
    )
    def test_find_ad_user_ad_dump(self, ad_dump, expected_exception):
        cpr = "112233-4455"

        def settings_transformer(settings):
            settings["integrations.ad"][0].update({"cpr_field": "cpr"})
            return settings

        self._setup_adwriter(
            early_transform_settings=settings_transformer, mock_find_ad_user=False
        )

        if expected_exception:
            with self.assertRaises(expected_exception):
                self.ad_writer._find_ad_user(cpr, ad_dump=ad_dump)
        else:
            ad_user = self.ad_writer._find_ad_user(cpr, ad_dump=ad_dump)
            self.assertDictEqual(ad_user, {"cpr": "112233-4455"})

    def test_template_fails_on_undefined_variable(self):
        settings_transformer = dict_modifier(
            {
                "integrations.ad_writer.template_to_ad_fields": {
                    "Displayname": "{{ unknown_variable }}"
                },
            }
        )
        self._setup_adwriter(early_transform_settings=settings_transformer)
        with self.assertRaises(UndefinedError):
            self.ad_writer.sync_user(mo_uuid="mo-uuid", sync_manager=False)

    def test_fullnames_are_empty_when_constituants_are_empty(self):
        self._setup_adwriter()
        uuid = "some_uuid_here"

        def get_mo_values(firstname, surname, nickname_firstname, nickname_surname):
            user = {
                "uuid": "some_uuid_here",
                "navn": "some_name some_lastname",
                "efternavn": surname,
                "fornavn": firstname,
                "kaldenavn": "",
                "kaldenavn_fornavn": nickname_firstname,
                "kaldenavn_efternavn": nickname_surname,
                "cpr": "some_cpr",
            }
            mock_lora_cache = AttrDict(
                {
                    "users": {user["uuid"]: [user]},
                    "engagements": {
                        "engagement_uuid": [
                            {
                                "user": user["uuid"],
                                "primary_boolean": True,
                                "user_key": "some_userkey",
                                "job_function": "job_function_title_uuid",
                                "unit": "some_unit",
                                "uuid": "engagement_uuid",
                                "from_date": None,
                                "to_date": None,
                            }
                        ]
                    },
                    "classes": {"job_function_title_uuid": {"title": "some_job_title"}},
                    "units": {
                        "some_unit": [
                            {
                                "uuid": "some_unit",
                                "name": "some_unit_name",
                                "user_key": "some_unit_userkey",
                                "location": "some_unit_location",
                                "unit_type": "some_unit_type",
                                "level": "some_unit_level",
                                "parent": None,
                            }
                        ]
                    },
                    "addresses": {},
                    "it_connections": {
                        "it_uuid": [
                            {
                                "uuid": "it_connection_uuid",
                                "user": "user_uuid",
                                "unit": None,
                                "username": "opus_username",
                                "itsystem": "it_uuid",
                                "from_date": "1930-01-01",
                                "to_date": None,
                            }
                        ]
                    },
                }
            )
            self.ad_writer.helper = mock.MagicMock(spec=MoraHelper)
            self.ad_writer.lc = mock_lora_cache
            self.ad_writer.lc_historic = mock_lora_cache
            self.ad_writer.datasource = LoraCacheSource(
                mock_lora_cache,  # lc
                mock_lora_cache,  # lc_historic
                MockMORESTSource(from_date=None, to_date=None),
            )
            mo_values = self.ad_writer._read_ad_information_from_mo(uuid)
            return mo_values

        mo_values = get_mo_values("Ursula", "Uniknavn", "Anne", "Jensen")
        assert isinstance(mo_values, LazyDict)
        assert mo_values["name"] == ("Ursula", "Uniknavn")
        assert mo_values["full_name"] == "Ursula Uniknavn"

        assert mo_values["nickname"] == ("Anne", "Jensen")
        assert mo_values["full_nickname"] == "Anne Jensen"

        mo_values = get_mo_values("", "", "", "")
        assert isinstance(mo_values, LazyDict)
        assert mo_values["name"] == ("", "")
        assert mo_values["full_name"] == ""

        assert mo_values["nickname"] == ("", "")
        assert mo_values["full_nickname"] == ""

        # Test all fields can be read
        for key in mo_values.keys():
            mo_values[key]

    def test_create_user_can_create_manager(self):
        # Test what happens when passing `create_manager=True` to `create_user`
        self._setup_adwriter()
        # Remember SamAccountName of user to be created
        expected_sam_account_name = self.mo_values_func()["sam_account_name"]
        # Replace `_wait_for_replication` with do-nothing stub.
        self.ad_writer._wait_for_replication = lambda sam_account_name: None
        # Create AD user and then set manager
        status, actual_sam_account_name = self.ad_writer.create_user(
            mo_uuid="mo-user-uuid", create_manager=True
        )
        # Assert we ran a PowerShell command to set user's manager
        self.assertIn("Set-ADUser -Manager", self.ad_writer.scripts[-1])
        # Assert we received the proper result from `create_user`
        self.assertTrue(status)
        self.assertEqual(actual_sam_account_name, expected_sam_account_name)

    def test_create_user_non_empty_response_is_error(self):
        # Test what happens when `create_user` encounters a non-empty response
        # from `_run_ps_script`.
        self._setup_adwriter()
        # Replace `get_ad_user` with function which always returns empty
        # response (otherwise `create_user` will raise an exception before we
        # reach the code we want to test.)
        self.ad_writer.get_from_ad = lambda **kwargs: {}
        # Replace `_run_ps_script` with function returning non-empty response
        self.ad_writer._run_ps_script = lambda ps_script: {"not": "empty"}
        # Try to create AD user
        status, msg = self.ad_writer.create_user(
            mo_uuid="mo-user-uuid", create_manager=False
        )
        self.assertFalse(status)
        self.assertRegex(msg, "Create user failed,.*")

    def test_create_user_bails_early_if_no_engagements(self):
        # Test what happens when `create_user` encounters an empty MO user
        self._setup_adwriter()
        # Replace `read_ad_information_from_mo` with function returning None
        self.ad_writer.read_ad_information_from_mo = (
            lambda mo_uuid, create_manager: None
        )
        # Assert that `create_user` bails early with an exception
        with self.assertRaises(NoPrimaryEngagementException):
            self.ad_writer.create_user(mo_uuid="mo-user-uuid", create_manager=False)

    def test_create_user_bails_early_if_ad_user_exists(self):
        # Test what happens if `create_user` encounters an already existing AD
        # user (either by username or CPR number lookup.)

        # Avoid circular imports
        from .mocks import MockADParameterReader
        from .mocks import MockOnlyCPRADParameterReader

        def assert_adwriter_get_ad_user_raises(
            mock_ad_reader_class, expected_exception
        ):
            self._setup_adwriter(mock_ad_reader_class=mock_ad_reader_class)
            # Assert we bail early with the proper exception
            with self.assertRaises(expected_exception):
                self.ad_writer.create_user(mo_uuid=MO_UUID, create_manager=False)

        # Replace `ADWriter._reader` with mock `ADParameterReader` which always returns
        # an AD user for *any* `user` or `cpr` argument.
        # This means that `ADWriter._check_if_user_exists` will "see" a mock AD user
        # when calling `ADWriter._reader.read_user(user=...)` and bail early.
        assert_adwriter_get_ad_user_raises(
            MockADParameterReader, SamAccountNameNotUnique
        )

        # Replace `ADWriter._reader` with mock `ADParameterReader` which returns None
        # when called with the `user` keyword argument, and a mock AD user when called
        # with the `cpr` keyword argument
        # This means that `ADWriter._check_if_user_exists` will "see" a mock AD user
        # when calling `ADWriter._reader.read_user(cpr=...)` and bail early.
        assert_adwriter_get_ad_user_raises(
            MockOnlyCPRADParameterReader, CprNotNotUnique
        )

    def test_other_attributes_skip_empty(self):
        # Configure template which tries to write `mo_values['foobar']` to the
        # 'fooBar' AD field.
        self._setup_adwriter(
            early_transform_settings=self._add_to_template_to_ad_fields(
                "fooBar",
                "{{ mo_values['foobar'] }}",
            ),
        )

        # Build `mo_values` containing an *empty* 'foobar' key
        mo_values = self._prepare_mo_values(
            mo_values_transformer=mo_modifier({"foobar": None})
        )

        # Render "New-ADUser" command using `mo_values`
        ps_script = self.ad_writer._get_create_user_command(mo_values, "sam")

        # Assert the command does not try to set "fooBar"="None"
        self.assertNotIn('"fooBar"="None"', ps_script)

    @parameterized.expand(
        [
            # Case 1: test that the old "search_base" setting is used if the
            # new setting "new_ad_user_path" is not provided.
            (
                {
                    "integrations.ad": [
                        {
                            "search_base": "custom_search_path",
                            # We need to provide these to make
                            # `_read_global_settings` pass.
                            "system_user": "system_user",
                            "password": "password",
                            # We need to provide this to make
                            # `_read_primary_ad_settings` pass.
                            "cpr_field": "cpr_field",
                        }
                    ]
                },
                "custom_search_path",
            ),
            # Case 2: test that the new "new_ad_user_path" settings is used, if
            # it is provided.
            (
                {"integrations.ad_writer.new_ad_user_path": "custom_ad_path"},
                "custom_ad_path",
            ),
        ]
    )
    def test_new_ad_user_path_setting_is_used(self, settings, expected_path_argument):
        self._setup_adwriter(early_transform_settings=dict_modifier(settings))

        # Test `_get_new_ad_user_path_argument`
        path_argument = self.ad_writer._get_new_ad_user_path_argument()
        self.assertRegex(path_argument, f' -Path "{expected_path_argument}"')

        # Test `_get_create_user_command` - the `ps_script` must end with a
        # "-Path" argument matching our expected path argument.
        mock_mo_user = mock.MagicMock()
        ps_script = self.ad_writer._get_create_user_command(
            mock_mo_user, "sam_account_name"
        )
        self.assertRegex(ps_script, f'.* -Path "{expected_path_argument}"$')


class _TestRealADWriter(TestCase):
    @staticmethod
    def mock_run_ps_response(status_code=0, std_out="{}", std_err=None):
        run_ps_response = MagicMock()
        run_ps_response.status_code = status_code
        run_ps_response.std_out = std_out
        run_ps_response.std_err = std_err
        return run_ps_response

    @staticmethod
    def _get_from_ad_matching_nothing(user=None, cpr=None, server=None):
        # Provide a mock implementation of `ADWriter.get_from_ad` which returns nothing
        return {}

    @staticmethod
    def _get_from_ad_matching_manager(user=None, cpr=None, server=None):
        # Provide a mock implementation of `ADWriter.get_from_ad` which can look up the
        # manager's 'SamAccountName'.
        employee_ad_user = {
            "SamAccountName": "user_sam",
            "Manager": None,
        }
        manager_ad_user = {
            "SamAccountName": MO_MANAGER_SAM,
            "DistinguishedName": "manager-dn",
        }
        if cpr == MO_USER_CPR:  # == the MO employee's own CPR
            return [employee_ad_user]
        if cpr == MO_MANAGER_CPR:
            return [manager_ad_user]

    def _prepare_adwriter(self, **kwargs):
        template_to_ad_fields = kwargs.pop("template_to_ad_fields", {})
        template_to_ad_fields_when_disable = kwargs.pop(
            "template_to_ad_fields_when_disable", {}
        )
        skip_locations = kwargs.pop("skip_locations", None)
        read_ou_addresses = kwargs.pop("read_ou_addresses", None)
        get_from_ad = kwargs.pop("get_from_ad", self._get_from_ad_matching_nothing)
        with MockADWriterContext(
            template_to_ad_fields=template_to_ad_fields,
            template_to_ad_fields_when_disable=template_to_ad_fields_when_disable,
            skip_locations=skip_locations,
            read_ou_addresses=read_ou_addresses,
            run_ps_response=kwargs.pop("run_ps_response", None),
            use_future_managers=kwargs.pop("use_future_managers", True),
        ):
            instance = ADWriter(**kwargs)
            instance.get_from_ad = get_from_ad
            return instance


class TestInitNameCreator(_TestRealADWriter):
    """Test `ADWriter._init_name_creator`.

    In this test, we instantiate the 'real' `ADWriter` rather than the
    `ADWriterTestSubclass` used in `TestADWriter`. This is because we want to
    test the behavior of the real `ADWriter._init_name_creator` method, rather
    than the overridden method in `ADWriterTestSubclass`.
    """

    def test_init_name_creator_reads_ad_usernames(self):
        """Calling `ADWriter` with the default args should automatically load
        occupied usernames from AD.
        """
        ad_writer = self._prepare_adwriter()
        # Assert that the single AD user mocked by `MockADParameterReader` is
        # loaded as an occupied name.
        self.assertGreater(len(ad_writer.name_creator.occupied_names), 0)
        # Assert that we loaded the usernames from AD.
        self.assertEqual(len(ad_writer.name_creator._loaded_occupied_name_sets), 1)
        self.assertIsInstance(
            ad_writer.name_creator._loaded_occupied_name_sets[0], UserNameSetInAD
        )

    def test_init_name_creator_skips_ad_usernames(self):
        """Calling `ADWriter` with `skip_occupied_names=True` should skip
        loading occupied usernames from AD as well as other sources of occupied
        usernames.
        """
        ad_writer = self._prepare_adwriter(skip_occupied_names=True)
        # Assert that no occupied usernames were loaded, and no username sets
        # were instantiated.
        self.assertSetEqual(ad_writer.name_creator.occupied_names, set())
        self.assertEqual(len(ad_writer.name_creator._loaded_occupied_name_sets), 0)


class TestSyncCompare(_TestRealADWriter):
    _ad_user_employee = {
        "cpr_field": MO_USER_CPR,
        "SamAccountName": MO_USER_SAM,
        "Manager": "old_manager_dn",
    }

    def test_compare_fields_converts_ad_list(self):
        """Some AD fields contain one-element lists, rather than the usual strings,
        numbers or UUIDs.

        In such cases, we "unpack" the single-element list before comparing it to the
        corresponding MO value - otherwise the comparison will not work as expected.

        See: #47148
        """
        # Arrange
        mo_value = "mo_value"
        ad_list_element = "ad_list_element"
        ad_user = {"ad_field_name": [ad_list_element]}
        ad_writer = self._prepare_adwriter()
        # Act
        mismatch = ad_writer._compare_fields("ad_field_name", mo_value, ad_user)
        # Assert
        self.assertIn("ad_field_name", mismatch)
        self.assertEqual(mismatch["ad_field_name"], (ad_list_element, mo_value))

    @parameterized.expand(
        [
            # 1. The MO value is None, and the AD value is None - not a mismatch
            (None, None, {}),
            # 2. The MO value is "None", and the AD value is None - not a mismatch
            ("None", None, {}),
            # 3. The MO value is None, and the AD value is "None" - not a mismatch
            (None, "None", {}),
            # 4. The MO value is "None", and the AD value is "None" - not a mismatch
            ("None", "None", {}),
            # 5. Neither MO nor AD is None or "None" - not a mismatch
            ("not none", "not none", {}),
            # 6. The MO value is None, but the AD value is not None - a mismatch
            (None, "not none", {"ad_field_name": ("not none", None)}),
            # 7. The MO value is "None", but the AD value is not None - a mismatch
            ("None", "not none", {"ad_field_name": ("not none", None)}),
            # 8. The MO value is not None, but the AD value is None - a mismatch
            ("not none", None, {"ad_field_name": (None, "not none")}),
            # 9. The MO value is not None, but the AD value is "None" - a mismatch
            ("not none", "None", {"ad_field_name": (None, "not none")}),
        ]
    )
    def test_compare_fields_handles_none(self, mo_value, ad_value, expected_mismatch):
        """When comparing the MO value `"None"` to the AD value `None`, do not consider
        the AD and MO fields to differ - otherwise we will make a lot of pointless
        updates in AD, giving empty AD fields the value `"None"`.

        Even if no other AD fields need to be updated, this will cause `mo_to_ad_sync`
        to update every AD user, causing the program to take a very long time to finish.

        Also, consider the reverse situation (AD value is `"None"`, MO value is `None`)
        as a non-mismatch.

        See: #50291
        """
        # Arrange
        ad_user = {"ad_field_name": ad_value}
        ad_writer = self._prepare_adwriter()
        # Act
        mismatch = ad_writer._compare_fields("ad_field_name", mo_value, ad_user)
        # Assert
        self.assertEqual(mismatch, expected_mismatch)

    @parameterized.expand(
        [
            (
                [],  # manager not present in AD
                None,  # expected `mismatch["manager"]`
                "could not find manager by cpr",  # expected logging
            ),
            (
                # manager present in AD
                [{"cpr_field": "manager_cpr", "DistinguishedName": "manager_dn"}],
                ("old_manager_dn", "manager_dn"),  # expected `mismatch["manager"]`
                "Manager should be updated",  # expected logging
            ),
        ]
    )
    def test_handles_manager_cpr_not_in_ad_dump(
        self, ad_dump_extra, expected_manager_mismatch, expected_log_message
    ):
        """When `ADWriter.sync_user` is called from `run_mo_to_ad_sync`, an `ad_dump`
        is provided. However, sometimes `sync_user` cannot find the AD user specified
        by `mo_values["manager_cpr"] in `ad_dump` and crashes unexpectedly.

        This test verifies that we no longer crash in such cases, but simply do not
        update the `Manager` attribute in AD.

        See: #50160
        """
        ad_writer = self._prepare_adwriter()
        ad_dump = [
            # employee AD user
            self._ad_user_employee,
            # 0 or 1 manager AD users
            *ad_dump_extra,
        ]
        mo_values = {
            "cpr": MO_USER_CPR,
            "full_name": "Full Name",
            "manager_cpr": "manager_cpr",
        }
        with self.assertLogs() as actual_log_messages:
            mismatch = ad_writer._sync_compare(mo_values, ad_dump=ad_dump)

        # Assert we got the expected `mismatch["manager"]` value
        actual_manager_mismatch = mismatch.get("manager")
        self.assertEqual(actual_manager_mismatch, expected_manager_mismatch)

        # Assert we logged the expected message
        self.assertEqual(actual_log_messages.records[-1].message, expected_log_message)

    def test_sync_compare_skips_invalid_mo_values(self):
        ad_writer = self._prepare_adwriter(
            # Map MO unit address to a hypothetical AD field called `ad_field_name`
            template_to_ad_fields={"ad_field_name": "{{ mo_values['unit_city'] }}"},
            # Ensure that MO returns an invalid unit address (DAR offline scenario)
            read_ou_addresses={"Adresse": "Ukendt"},
        )
        ad_dump = [self._ad_user_employee]
        mo_values = ad_writer.read_ad_information_from_mo("uuid")

        with self.assertLogs() as actual_log_messages:
            mismatch = ad_writer._sync_compare(mo_values, ad_dump=ad_dump)

        # Assert that we did not consider the MO and AD values to be different and thus
        # need synchronization.
        self.assertNotIn("ad_field_name", mismatch)

        # Assert that we logged the expected message
        self.assertEqual(
            # Find matching log message
            only(
                record.message
                for record in actual_log_messages.records
                if "INVALID" in record.message
            ),
            # Expected content
            "'ad_field_name': MO value is INVALID, not changing AD value None",
        )

        # Assert that the generated PowerShell script does not include the invalid value
        with mock.patch.object(ad_writer, "_run_ps_script") as mock_run_ps_script:
            ad_writer.sync_user("mo_uuid", ad_dump=ad_dump)
            ps_script = mock_run_ps_script.call_args[0][0]
            self.assertNotIn(f'"ad_field_name"="{INVALID}"', ps_script)


class TestPreview(_TestRealADWriter):
    def test_preview_create_command(self):
        ad_writer = self._prepare_adwriter()
        create_cmds = ad_writer._preview_create_command(MO_UUID)
        self.assertEqual(len(create_cmds), 2)
        self.assertIn("New-ADUser", create_cmds[0])
        self.assertIn("Set-ADUser -Manager", create_cmds[1])

    def test_preview_sync_command_can_preview_rename_when_ad_dump_is_none(self):
        ad_writer = self._prepare_adwriter(
            get_from_ad=self._get_from_ad_matching_nothing
        )
        (
            sync_cmd,
            rename_cmd,
            rename_cmd_target,
            add_manager_cmd,
        ) = ad_writer._preview_sync_command(MO_UUID, "user_sam")
        # Examine 'sync_cmd'
        self.assertIn("Get-ADUser", sync_cmd)
        self.assertIn("Set-ADUser", sync_cmd)
        # Examine 'rename_cmd'
        self.assertIn("Get-ADUser", rename_cmd)
        self.assertIn("Rename-ADobject", rename_cmd)
        self.assertIn('-NewName "<new name>"', rename_cmd)
        self.assertEqual("<nonexistent AD user>", rename_cmd_target)
        # Examine 'add_manager_cmd'
        self.assertEqual("", add_manager_cmd)

    def test_preview_sync_command_can_preview_rename(self):
        reader = MockADParameterReaderWithManager()
        lc = MockLoraCacheWithManager()
        mo_name = lc._mo_values_employee["navn"]

        # Mock `ad_dump` which only contains an employee, and no manager
        ad_dump = [reader.read_user(cpr=MO_USER_CPR)]

        with MockADWriterContext(
            template_to_ad_fields={"Name": "{{ mo_values['full_name'] }}"}
        ):
            writer = ADWriter(lc=lc, lc_historic=lc)
            (
                sync_cmd,
                rename_cmd,
                rename_cmd_target,
                add_manager_cmd,
            ) = writer._preview_sync_command(MO_UUID, "user_sam", ad_dump=ad_dump)
            # Examine 'sync_cmd'
            self.assertIn("Get-ADUser", sync_cmd)
            self.assertIn("Set-ADUser", sync_cmd)
            # Examine 'rename_cmd'
            self.assertIn("Get-ADUser", rename_cmd)
            self.assertIn("Rename-ADobject", rename_cmd)
            self.assertIn(f'-NewName "{mo_name}"', rename_cmd)
            self.assertIsNone(rename_cmd_target)
            # Examine 'add_manager_cmd'
            self.assertEqual("", add_manager_cmd)

    def test_preview_sync_command_can_preview_add_manager(self):
        reader = MockADParameterReaderWithManager()
        lc = MockLoraCacheWithManager()
        ad_dump = [
            reader.read_user(cpr=MO_USER_CPR),  # employee
            reader.read_user(cpr=MO_MANAGER_CPR),  # manager
        ]
        with MockADWriterContext():
            writer = ADWriter(lc=lc, lc_historic=lc)
            (
                sync_cmd,
                rename_cmd,
                rename_cmd_target,
                add_manager_cmd,
            ) = writer._preview_sync_command(MO_UUID, "user_sam", ad_dump=ad_dump)
            # Examine 'sync_cmd'
            self.assertIn("Get-ADUser", sync_cmd)
            self.assertIn("Set-ADUser", sync_cmd)
            # Examine 'rename_cmd'
            self.assertEqual("", rename_cmd)
            # Examine 'add_manager_cmd'
            self.assertIn("Get-ADUser", add_manager_cmd)
            self.assertIn("Set-ADUser", add_manager_cmd)
            self.assertIn(f"-Manager {MO_MANAGER_SAM}", add_manager_cmd)

    @parameterized.expand(
        [
            # 1. Test disabling AD user without additional field mapping.
            (
                # `enable` argument
                False,
                # expected value of `-Enable` argument in PowerShell command
                "$false",
                # additional field mapping
                None,
                # expected output from additional field mapping
                None,
            ),
            # 2. Test disabling AD user with an additional field mapping.
            (
                # `enable` argument
                False,
                # expected value of `-Enable` argument in PowerShell command
                "$false",
                # additional field mapping
                {"extensionAttribute": "{{ now.strftime('%Y-%m-%d') }}"},
                # expected output from additional field mapping
                '"extensionAttribute"="%s"' % datetime.now().strftime("%Y-%m-%d"),
            ),
            # 3. Test enabling AD user without additional field mapping.
            (
                # `enable` argument
                True,
                # expected value of `-Enable` argument in PowerShell command
                "$true",
                # additional field mapping
                None,
                # expected output from additional field mapping
                None,
            ),
            # 4. Test enabling AD user with an additional field mapping.
            # Since we are enabling the AD user, the additional field mapping should
            # not be used.
            (
                # `enable` argument
                True,
                # expected value of `-Enable` argument in PowerShell command
                "$true",
                # additional field mapping
                {"extensionAttribute": "{{ now.strftime('%Y-%m-%d') }}"},
                # expected output from additional field mapping
                None,
            ),
        ]
    )
    def test_preview_enable_command(
        self,
        enable_arg: bool,
        expected_enabled_arg_value: str,
        additional_field_mapping: dict,
        expected_additional_output: str,
    ):
        ad_writer_context_kwargs = (
            {"template_to_ad_fields_when_disable": additional_field_mapping}
            if additional_field_mapping
            else {}
        )
        ad_writer = self._prepare_adwriter(**ad_writer_context_kwargs)
        enable_cmd = ad_writer._get_enable_user_cmd("sam_account_name", enable_arg)
        self.assertIn("Get-ADUser", enable_cmd)
        self.assertIn("Set-ADUser", enable_cmd)
        self.assertIn("-Enabled %s" % expected_enabled_arg_value, enable_cmd)
        if expected_additional_output:
            self.assertIn(expected_additional_output, enable_cmd)


class TestReadADInformationFromMO(_TestRealADWriter):
    @parameterized.expand(
        [
            # When DAR lookups work, MO returns addresses on the format
            # "Testvej 123, 1234 Testby".
            # Check that we parse such values correctly into its constituent parts.
            (
                {"Adresse": "Testvej 123, 1234 Testby"},
                {"postal_code": "1234", "city": "Testby", "streetname": "Testvej 123"},
            ),
            # When DAR is offline, MO returns addresses with the literal value "Ukendt".
            # We convert them into invalid unit addresses.
            (
                {"Adresse": "Ukendt"},
                ADWriter.INVALID_UNIT_ADDRESS,
            ),
            # If we get an invalid Danish postal address, we also convert them into an
            # invalid unit address.
            (
                {"Adresse": "Ikke en gyldig dansk postadresse"},
                ADWriter.INVALID_UNIT_ADDRESS,
            ),
            # Empty MO response causes an invalid unit address as well.
            (
                {},
                ADWriter.INVALID_UNIT_ADDRESS,
            ),
        ]
    )
    def test_parsed_addresses(self, address, expected_parsed_address):
        ad_writers = (
            ("ADWriter using MO API", self._get_mo_ad_writer(address)),
            ("ADWriter using LoraCache", self._get_loracache_ad_writer(address)),
        )
        for name, ad_writer in ad_writers:
            with self.subTest(name):
                mo_values = ad_writer.read_ad_information_from_mo(MO_UUID)
                self.assertEqual(
                    mo_values["_parsed_addresses"], expected_parsed_address
                )
                if expected_parsed_address is ADWriter.INVALID_UNIT_ADDRESS:
                    self.assertEqual(mo_values["unit_postal_code"], INVALID)
                    self.assertEqual(mo_values["unit_city"], INVALID)
                    self.assertEqual(mo_values["unit_streetname"], INVALID)

    def _get_mo_ad_writer(self, address):
        return self._prepare_adwriter(read_ou_addresses=address)

    def _get_loracache_ad_writer(self, address):
        lc = MockLoraCacheUnitAddress(address_value=address.get("Adresse"))
        return self._prepare_adwriter(lc=lc, lc_historic=lc)


class TestEnableUser(_TestRealADWriter):
    @parameterized.expand(
        [
            (False, "disabled AD user"),
            (True, "enabled AD user"),
        ]
    )
    def test_enable_user_handles_ok_response(self, enable: bool, expected_msg: str):
        # Arrange: mock an "OK" response from WinRM
        ad_writer = self._prepare_adwriter(run_ps_response=self.mock_run_ps_response())
        # Act
        result = ad_writer.enable_user("sam_account_name", enable)
        # Assert (here `True` means the PowerShell command executed successfully)
        self.assertEqual(result, (True, expected_msg))

    def test_enable_user_handles_error_response(self):
        # Arrange: mock an "error" response from WinRM
        ad_writer = self._prepare_adwriter(
            run_ps_response=self.mock_run_ps_response(
                status_code=1,
                std_err="something wrong",
            )
        )
        # Act and assert
        with self.assertRaises(CommandFailure):
            ad_writer.enable_user("sam_account_name", False)

    def test_get_enable_user_cmd_strips_linebreaks(self):
        ad_writer = self._prepare_adwriter()
        ps_script = ad_writer._get_enable_user_cmd("user", enable=False)
        # Find actual "Get-ADUser NNN | Set-ADUser NNN -Enabled $false" line in PS
        match = re.search("Get-ADUser .*", ps_script, re.MULTILINE)
        # Assert that this line does not contain any unintended line breaks, caused by
        # forgetting to call `AD.remove_redundant`.
        assert "\n" not in match.group()


class TestSkipLocation(_TestRealADWriter):
    _combinations = [
        # 1. If `skip_locations` is not set, expect `mo_values` to be present, and
        # contain the expected location.
        (
            # Value of the `skip_locations` setting
            None,
            # Assert that `mo_values` is returned and contains the expected location
            lambda mo_values: mo_values["location"] == MO_ROOT_ORG_UNIT_NAME,
        ),
        # 2. If `skip_locations` is set, but does not match the location in `mo_values`,
        # expect `mo_values` to be present and contain the expected location.
        (
            ["Something that is not MO_ROOT_ORG_UNIT_NAME"],
            lambda mo_values: mo_values["location"] == MO_ROOT_ORG_UNIT_NAME,
        ),
        # 3. If `skip_locations` is set, and matches the location in `mo_values`,
        # expect `mo_values` to be None, indicating that this MO user should be skipped.
        (
            [MO_ROOT_ORG_UNIT_NAME],
            lambda mo_values: mo_values is None,
        ),
    ]

    @parameterized.expand(_combinations)
    def test_skip_location_mo(self, skip_locations, test_condition):
        """Test `ADWriter.read_ad_information_from_mo` when reading from MO"""
        ad_writer = self._prepare_adwriter(skip_locations=skip_locations)
        mo_values = ad_writer.read_ad_information_from_mo(MO_UUID)
        assert test_condition(mo_values)

    @parameterized.expand(_combinations)
    def test_skip_location_loracache(self, skip_locations, test_condition):
        """Test `ADWriter.read_ad_information_from_mo` when reading from LoraCache"""
        lc = MockLoraCacheUnitAddress(address_value=None)
        ad_writer = self._prepare_adwriter(
            skip_locations=skip_locations, lc=lc, lc_historic=lc
        )
        mo_values = ad_writer.read_ad_information_from_mo(MO_UUID)
        assert test_condition(mo_values)

    @given(
        st.lists(st.text()) | st.none(),
        st.builds(lambda xs: "\\".join(xs), st.lists(st.text())) | st.none(),
    )
    def test_skip_unit(self, skip_locations, location):
        ad_writer = self._prepare_adwriter(skip_locations=skip_locations)
        result = ad_writer._skip_unit({"location": location})
        if not skip_locations:
            assert result is False
        elif location and any(part in skip_locations for part in location.split("\\")):
            assert result is True
        else:
            assert result is None


class TestRenameADUser(_TestRealADWriter):
    def test_sleep(self):
        # Arrange: mock a successful WinRM invocation
        ad_writer = self._prepare_adwriter(run_ps_response=self.mock_run_ps_response())

        with mock.patch("time.sleep") as mock_time_sleep:
            # Act
            ad_writer._rename_ad_user("user_sam", "New Name")
            # Assert
            mock_time_sleep.assert_called_once_with(1)


class TestGetManagerUUID(_TestRealADWriter):
    def test_get_manager_uuid(self):
        ad_writer = self._prepare_adwriter(
            get_from_ad=self._get_from_ad_matching_manager
        )
        mo_values = ad_writer.read_ad_information_from_mo(MO_UUID)
        assert mo_values["_manager_uuid"] == MO_MANAGER_UUID
        assert mo_values["_manager_mo_user"]["uuid"] == MO_MANAGER_UUID
        assert mo_values["read_manager"] is True
        assert mo_values["manager_name"] == mo_values["_manager_mo_user"]["name"]
        assert mo_values["manager_cpr"] == mo_values["_manager_mo_user"]["cpr_no"]
        assert mo_values["manager_mail"] == "address-value"
        assert mo_values["manager_sam"] == MO_MANAGER_SAM


class TestUseFutureManagersFeatureFlag:
    def test_effect_when_false(self, caplog):
        with MockADWriterContext(use_future_managers=False):
            with caplog.at_level(logging.INFO):
                ad_writer = ADWriter()
                # Assert that `ADWriter._use_graphql_source_if_feature_flagged` does
                # not monkeypatch `self.datasource.get_manager_uuid`.
                assert inspect.getsourcelines(
                    ad_writer.datasource.get_manager_uuid
                ) == inspect.getsourcelines(MockMORESTSource.get_manager_uuid)
                assert "Not using MOGraphqlSource" in caplog.text

    def test_effect_when_true(self, caplog):
        with MockADWriterContext(use_future_managers=True):
            with caplog.at_level(logging.INFO):
                ad_writer = ADWriter()
                # Assert that `ADWriter._use_graphql_source_if_feature_flagged`
                # monkeypatches `self.datasource.get_manager_uuid` when True.
                assert inspect.getsourcelines(
                    ad_writer.datasource.get_manager_uuid
                ) == inspect.getsourcelines(MockMOGraphqlSource.get_manager_uuid)
                assert "Using MOGraphqlSource to patch" in caplog.text


class TestUpdateManager(_TestRealADWriter):
    # Mock an AD with an employee and a manager. The employee does not yet have its
    # `Manager` field set, so `ADWriter.sync_user` should try to update it.
    _ad_dump = [
        # Employee
        {
            "cpr_field": MO_USER_CPR,
            "SamAccountName": MO_USER_SAM,
            "Manager": None,
        },
        # Manager
        {
            "cpr_field": MO_MANAGER_CPR,
            "SamAccountName": MO_MANAGER_SAM,
            "DistinguishedName": "foo",
        },
    ]

    @parameterized.expand(
        [
            # Case 1: `ADWriter.datasource.get_manager_uuid` returns a manager UUID for
            # the employee in question. The AD user should be updated.
            (
                # Manager UUID
                MO_MANAGER_UUID,
                # Assertion
                lambda mock_add_manager: mock_add_manager.assert_called_once_with(
                    user_sam=MO_USER_SAM,
                    manager_sam=MO_MANAGER_SAM,
                ),
            ),
            # Case 2: `ADWriter.datasource.get_manager_uuid` returns None for the
            # employee in question. The AD user should *not* be updated.
            (
                # Manager UUID
                None,
                # Assertion
                lambda mock_add_manager: mock_add_manager.assert_not_called(),
            ),
        ]
    )
    def test_sync_user_calls_add_manager_to_user(self, mo_manager_uuid, assertion):
        ad_writer = self._prepare_adwriter(run_ps_response=self.mock_run_ps_response())
        with mock.patch.object(
            ad_writer.datasource,
            "get_manager_uuid",
            return_value=mo_manager_uuid,
        ):
            with mock.patch.object(
                ad_writer, "add_manager_to_user"
            ) as mock_add_manager:
                ad_writer.sync_user(MO_UUID, ad_dump=self._ad_dump, sync_manager=True)
                assertion(mock_add_manager)



import datetime
import logging
import typing
from typing import Tuple
from typing import Type
from typing import TypeVar
from uuid import UUID

import click
import ra_utils.ensure_single_run
from alembic.migration import MigrationContext
from alembic.operations import Operations
from more_itertools import ichunked
from more_itertools import one
from ra_utils.ensure_single_run import ensure_single_run
from ra_utils.job_settings import JobSettings
from ra_utils.load_settings import load_settings
from ra_utils.tqdm_wrapper import tqdm
from sqlalchemy import create_engine
from sqlalchemy import select
from sqlalchemy.engine import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

from .gql_lora_cache_async import GQLLoraCache
from .lora_cache import get_cache as LoraCache
from .sql_table_defs import Adresse
from .sql_table_defs import Base
from .sql_table_defs import Bruger
from .sql_table_defs import DARAdresse
from .sql_table_defs import Engagement
from .sql_table_defs import Enhed
from .sql_table_defs import Enhedssammenkobling
from .sql_table_defs import Facet
from .sql_table_defs import ItForbindelse
from .sql_table_defs import ItSystem
from .sql_table_defs import Klasse
from .sql_table_defs import KLE
from .sql_table_defs import Kvittering
from .sql_table_defs import Leder
from .sql_table_defs import LederAnsvar
from .sql_table_defs import Orlov
from .sql_table_defs import sql_type
from .sql_table_defs import Tilknytning
from .sql_table_defs import WAdresse
from .sql_table_defs import WBruger
from .sql_table_defs import WDARAdresse
from .sql_table_defs import WEngagement
from .sql_table_defs import WEnhed
from .sql_table_defs import WEnhedssammenkobling
from .sql_table_defs import WFacet
from .sql_table_defs import WItForbindelse
from .sql_table_defs import WItSystem
from .sql_table_defs import WKlasse
from .sql_table_defs import WKLE
from .sql_table_defs import WLeder
from .sql_table_defs import WLederAnsvar
from .sql_table_defs import WOrlov
from .sql_table_defs import WTilknytning
from .sql_url import DatabaseFunction
from .sql_url import generate_connection_url
from .sql_url import generate_engine_settings


_T_Facet = TypeVar("_T_Facet", Facet, WFacet)
_T_Klasse = TypeVar("_T_Klasse", Klasse, WKlasse)
_T_Bruger = TypeVar("_T_Bruger", Bruger, WBruger)
_T_Enhed = TypeVar("_T_Enhed", Enhed, WEnhed)
_T_Adresse = TypeVar("_T_Adresse", Adresse, WAdresse)
_T_Engagement = TypeVar("_T_Engagement", Engagement, WEngagement)
_T_Tilknytning = TypeVar("_T_Tilknytning", Tilknytning, WTilknytning)
_T_Orlov = TypeVar("_T_Orlov", Orlov, WOrlov)
_T_ItSystem = TypeVar("_T_ItSystem", ItSystem, WItSystem)
_T_ItForbindelse = TypeVar("_T_ItForbindelse", ItForbindelse, WItForbindelse)
_T_Leder = TypeVar("_T_Leder", Leder, WLeder)
_T_LederAnsvar = TypeVar("_T_LederAnsvar", LederAnsvar, WLederAnsvar)
_T_KLE = TypeVar("_T_KLE", KLE, WKLE)
_T_Enhedssammenkobling = TypeVar(
    "_T_Enhedssammenkobling", Enhedssammenkobling, WEnhedssammenkobling
)
_T_DARAdresse = TypeVar("_T_DARAdresse", DARAdresse, WDARAdresse)


class SqlExportSettings(JobSettings):
    class Config:
        settings_json_prefix = "exporters.actual_state"


logger = logging.getLogger(__name__)


class SqlExport:
    def __init__(self, force_sqlite=False, historic=False, settings=None):
        logger.info("Start SQL export")
        self.force_sqlite = force_sqlite
        self.historic = historic
        self.settings = settings
        self.engine = self._get_engine()
        self.export_cpr = self._get_export_cpr_setting()
        self.chunk_size = 5000
        self.lc = None

    def _get_engine(self) -> Engine:
        database_function = DatabaseFunction.ACTUAL_STATE
        if self.historic:
            database_function = DatabaseFunction.ACTUAL_STATE_HISTORIC
        db_string = generate_connection_url(
            database_function, force_sqlite=self.force_sqlite, settings=self.settings
        )
        engine_settings = generate_engine_settings(
            database_function, force_sqlite=self.force_sqlite, settings=self.settings
        )
        return create_engine(db_string, **engine_settings)

    def _get_export_cpr_setting(self) -> bool:
        return self.settings.get("exporters.actual_state.export_cpr", True)

    def _get_lora_cache(self, resolve_dar, use_pickle) -> GQLLoraCache:
        if self.historic:
            lc = LoraCache(
                resolve_dar=resolve_dar, full_history=True, settings=self.settings
            )
            lc.populate_cache(dry_run=use_pickle)
        else:
            lc = LoraCache(resolve_dar=resolve_dar, settings=self.settings)
            lc.populate_cache(dry_run=use_pickle)
            lc.calculate_derived_unit_data()
            lc.calculate_primary_engagements()
        return lc

    def _get_db_session(self) -> Session:
        Session = sessionmaker(bind=self.engine, autoflush=False)
        return Session()

    def _get_lora_class(self, uuid: str) -> Tuple[str, dict]:
        cls: dict = self.lc.classes.get(uuid) or {"title": uuid}
        return uuid, cls

    def perform_export(self, resolve_dar=True, use_pickle=None):
        def timestamp():
            return datetime.datetime.now()

        tables = dict(Base.metadata.tables)

        logger.info("Dropping work tables")
        Base.metadata.drop_all(
            self.engine,
            tables=[table for name, table in tables.items() if name[0] == "w"],
        )
        logger.info("Ensure work tables and 'kvittering' exists")
        Base.metadata.create_all(
            self.engine,
            tables=[
                table
                for name, table in tables.items()
                if name[0] == "w" or name == "kvittering"
            ],
        )

        self.session = self._get_db_session()

        query_time = timestamp()
        kvittering = self._add_receipt(query_time)
        self.lc = self.lc or self._get_lora_cache(resolve_dar, use_pickle)

        start_delivery_time = timestamp()
        self._update_receipt(kvittering, start_delivery_time)

        tasks = [
            self._add_facets,
            self._add_classes,
            self._add_units,
            self._add_users,
            self._add_addresses,
            self._add_dar_addresses,
            self._add_engagements,
            self._add_associations,
            self._add_leaves,
            self._add_managers,
            self._add_it_systems,
            self._add_it_users,
            self._add_kles,
            self._add_related,
        ]
        for task in tqdm(tasks, desc="SQLExport", unit="task"):
            task()

        end_delivery_time = timestamp()
        self._update_receipt(kvittering, start_delivery_time, end_delivery_time)

    def get_actual_tables(self):
        connection = self.engine.connect()
        inspector = Inspector.from_engine(connection)
        actual_tables = inspector.get_table_names()
        return set(actual_tables)

    def swap_tables(self):
        """Swap tables around to present the exported data.

        Swaps the current tables to old tables, then swaps write tables to current.
        Finally drops the old tables leaving just the current tables.
        """
        logger.info("Swapping tables")
        connection = self.engine.connect()
        ctx = MigrationContext.configure(connection)
        op = Operations(ctx)

        def gen_table_names(write_table):
            """Generate current and old table names from write tables."""
            # Current tables do not have the prefix 'w'
            current_table = write_table[1:]
            old_table = current_table + "_old"
            return write_table, current_table, old_table

        tables = dict(Base.metadata.tables)
        tables = {t for t in tables if t[0] == "w"}
        tables = list(map(gen_table_names, tables))

        # Drop any left-over old tables that may exist
        with ctx.begin_transaction():
            actual_tables = self.get_actual_tables()
            for _, _, old_table in tables:
                if old_table in actual_tables:
                    op.drop_table(old_table)

        # Rename current to old and write to current
        with ctx.begin_transaction():
            actual_tables = self.get_actual_tables()
            for write_table, current_table, old_table in tables:
                if current_table in actual_tables:
                    op.rename_table(current_table, old_table)
                # Rename write table to current table
                op.rename_table(write_table, current_table)

        # Drop any old tables that may exist
        with ctx.begin_transaction():
            actual_tables = self.get_actual_tables()
            for _, _, old_table in tables:
                if old_table in actual_tables:
                    op.drop_table(old_table)

    def _generate_sql_facets(self, uuid, facet_info, model: Type[_T_Facet]) -> _T_Facet:
        return model(
            uuid=str(uuid),
            bvn=facet_info["user_key"],
        )

    def _add_facets(self) -> None:
        logger.info("Add classification")
        facets = tqdm(self.lc.facets.items(), desc="Export facet", unit="facet")
        for chunk in ichunked(facets, self.chunk_size):
            for uuid, facet_info in chunk:
                sql_facet = self._generate_sql_facets(uuid, facet_info, WFacet)
                self.session.add(sql_facet)
            self.session.commit()

    def _generate_sql_classes(
        self, uuid, klasse_info, model: Type[_T_Klasse]
    ) -> _T_Klasse:
        return model(
            uuid=str(uuid),
            bvn=klasse_info["user_key"],
            titel=klasse_info["title"],
            facet_uuid=klasse_info["facet"],
            facet_bvn=self.lc.facets[klasse_info["facet"]]["user_key"],
        )

    def _add_classes(self) -> None:
        classes = tqdm(self.lc.classes.items(), desc="Export class", unit="class")
        for chunk in ichunked(classes, self.chunk_size):
            for uuid, klasse_info in chunk:
                sql_class = self._generate_sql_classes(uuid, klasse_info, WKlasse)
                self.session.add(sql_class)
            self.session.commit()

    def _generate_sql_users(self, uuid, user_info, model: Type[_T_Bruger]) -> _T_Bruger:
        return model(
            uuid=str(uuid),
            bvn=user_info["user_key"],
            fornavn=user_info["fornavn"],
            efternavn=user_info["efternavn"],
            kaldenavn_fornavn=user_info["kaldenavn_fornavn"],
            kaldenavn_efternavn=user_info["kaldenavn_efternavn"],
            cpr=user_info["cpr"] if self.export_cpr else "",
            startdato=user_info["from_date"],
            slutdato=user_info["to_date"],
        )

    def _add_users(self) -> None:
        logger.info("Add users")
        users = tqdm(self.lc.users.items(), desc="Export user", unit="user")
        for chunk in ichunked(users, self.chunk_size):
            for uuid, user_effects in chunk:
                for user_info in user_effects:
                    sql_user = self._generate_sql_users(uuid, user_info, WBruger)
                    self.session.add(sql_user)
            self.session.commit()

    def _generate_sql_units(self, uuid, unit_info, model: Type[_T_Enhed]) -> _T_Enhed:
        location = unit_info.get("location")
        manager_uuid = unit_info.get("manager_uuid")
        acting_manager_uuid = unit_info.get("acting_manager_uuid")

        unit_type = unit_info["unit_type"]

        enhedsniveau_titel = ""
        if unit_info["level"]:
            enhedsniveau_titel = self.lc.classes[unit_info["level"]]["title"]

        tidsregistrering_titel = ""
        if unit_info["time_planning"]:
            tidsregistrering_titel = self.lc.classes[unit_info["time_planning"]][
                "title"
            ]

        (
            org_unit_hierarchy_uuid,
            org_unit_hierarchy_class,
        ) = self._get_lora_class(unit_info["org_unit_hierarchy"])

        return model(
            uuid=str(uuid),
            navn=unit_info["name"],
            bvn=unit_info["user_key"],
            forældreenhed_uuid=unit_info["parent"],
            enhedstype_uuid=unit_type,
            enhedsniveau_uuid=unit_info["level"],
            tidsregistrering_uuid=unit_info["time_planning"],
            organisatorisk_sti=location,
            leder_uuid=manager_uuid,
            fungerende_leder_uuid=acting_manager_uuid,
            enhedstype_titel=self.lc.classes[unit_type]["title"],
            enhedsniveau_titel=enhedsniveau_titel,
            tidsregistrering_titel=tidsregistrering_titel,
            opmærkning_uuid=org_unit_hierarchy_uuid,
            opmærkning_titel=org_unit_hierarchy_class["title"],
            startdato=unit_info["from_date"],
            slutdato=unit_info["to_date"],
        )

    def _add_units(self) -> None:
        logger.info("Add users")
        units = tqdm(self.lc.units.items(), desc="Export unit", unit="unit")
        for chunk in ichunked(units, self.chunk_size):
            for uuid, unit_validities in chunk:
                for unit_info in unit_validities:
                    sql_unit = self._generate_sql_units(uuid, unit_info, WEnhed)
                    self.session.add(sql_unit)
            self.session.commit()

    def _generate_sql_engagements(
        self, uuid, engagement_info, model: Type[_T_Engagement]
    ) -> _T_Engagement:
        if engagement_info["primary_type"] is not None:
            primærtype_titel = self.lc.classes[engagement_info["primary_type"]]["title"]
        else:
            primærtype_titel = ""

        engagement_type_uuid = engagement_info["engagement_type"]
        job_function_uuid, job_function_class = self._get_lora_class(
            engagement_info["job_function"]
        )

        return model(
            uuid=str(uuid),
            enhed_uuid=engagement_info["unit"],
            bruger_uuid=engagement_info["user"],
            bvn=engagement_info["user_key"],
            engagementstype_uuid=engagement_info["engagement_type"],
            primær_boolean=engagement_info.get("primary_boolean"),
            arbejdstidsfraktion=engagement_info["fraction"],
            engagementstype_titel=self.lc.classes.get(
                engagement_type_uuid, {"title": engagement_type_uuid}
            )["title"],
            primærtype_titel=primærtype_titel,
            stillingsbetegnelse_uuid=job_function_uuid,
            stillingsbetegnelse_titel=job_function_class["title"],
            primærtype_uuid=engagement_info["primary_type"],
            startdato=engagement_info["from_date"],
            slutdato=engagement_info["to_date"],
            **engagement_info["extensions"],
        )

    def _add_engagements(self) -> None:
        logger.info("Add engagements")
        engagements = tqdm(
            self.lc.engagements.items(), desc="Export engagement", unit="engagement"
        )
        for chunk in ichunked(engagements, self.chunk_size):
            for uuid, engagement_validity in chunk:
                for engagement_info in engagement_validity:
                    sql_engagement = self._generate_sql_engagements(
                        uuid, engagement_info, WEngagement
                    )
                    self.session.add(sql_engagement)
            self.session.commit()

    def _generate_sql_addresses(
        self, uuid, address_info, model: Type[_T_Adresse]
    ) -> _T_Adresse:
        visibility_text = None
        if address_info["visibility"] is not None:
            visibility_text = self.lc.classes[address_info["visibility"]]["title"]
        visibility_scope = None
        if address_info["visibility"] is not None:
            visibility_scope = self.lc.classes[address_info["visibility"]]["scope"]

        return model(
            uuid=str(uuid),
            enhed_uuid=address_info["unit"],
            bruger_uuid=address_info["user"],
            bvn=address_info["user_key"],
            værdi=address_info["value"],
            dar_uuid=address_info["dar_uuid"],
            adressetype_uuid=address_info["adresse_type"],
            adressetype_bvn=self.lc.classes[address_info["adresse_type"]]["user_key"],
            adressetype_scope=address_info["scope"],
            adressetype_titel=self.lc.classes[address_info["adresse_type"]]["title"],
            synlighed_uuid=address_info["visibility"],
            synlighed_scope=visibility_scope,
            synlighed_titel=visibility_text,
            startdato=address_info["from_date"],
            slutdato=address_info["to_date"],
        )

    def _add_addresses(self) -> None:
        logger.info("Add addresses")
        addresses = tqdm(
            self.lc.addresses.items(), desc="Export address", unit="address"
        )
        for chunk in ichunked(addresses, self.chunk_size):
            for uuid, address_validities in chunk:
                for address_info in address_validities:
                    sql_address = self._generate_sql_addresses(
                        uuid, address_info, WAdresse
                    )
                    self.session.add(sql_address)
            self.session.commit()

    def _generate_sql_dar_addresses(
        self, uuid, address_info, model: Type[_T_DARAdresse]
    ) -> _T_DARAdresse:
        return model(
            uuid=str(uuid),
            **{
                key: value
                for key, value in address_info.items()
                if key in model.__table__.columns.keys() and key not in ("id", "uuid")
            },
        )

    def _add_dar_addresses(self) -> None:
        logger.info("Add DAR addresses")
        dar = tqdm(self.lc.dar_cache.items(), desc="Export DAR", unit="DAR")
        for chunk in ichunked(dar, self.chunk_size):
            for uuid, address_info in chunk:
                sql_address = self._generate_sql_dar_addresses(
                    uuid, address_info, WDARAdresse
                )
                self.session.add(sql_address)
            self.session.commit()

    def _generate_sql_associations(
        self, uuid, association_info, model: Type[_T_Tilknytning]
    ) -> _T_Tilknytning:
        (
            association_type_uuid,
            association_type_class,
        ) = self._get_lora_class(association_info["association_type"])
        job_function_uuid, job_function_class = self._get_lora_class(
            association_info["job_function"]
        )
        return model(
            uuid=str(uuid),
            bruger_uuid=association_info["user"],
            enhed_uuid=association_info["unit"],
            bvn=association_info["user_key"],
            tilknytningstype_uuid=association_type_uuid,
            tilknytningstype_titel=association_type_class["title"],
            startdato=association_info["from_date"],
            slutdato=association_info["to_date"],
            it_forbindelse_uuid=association_info["it_user"],
            stillingsbetegnelse_uuid=job_function_uuid,
            stillingsbetegnelse_titel=job_function_class["title"],
            primær_boolean=association_info.get("primary_boolean"),
            faglig_organisation=association_info.get("dynamic_class"),
        )

    def _add_associations(self) -> None:
        logger.info("Add associations")
        associations = tqdm(
            self.lc.associations.items(), desc="Export association", unit="association"
        )
        for chunk in ichunked(associations, self.chunk_size):
            for uuid, association_validity in chunk:
                for association_info in association_validity:
                    sql_association = self._generate_sql_associations(
                        uuid, association_info, WTilknytning
                    )
                    self.session.add(sql_association)
            self.session.commit()

    def _generate_sql_leave(self, uuid, leave_info, model: Type[_T_Orlov]) -> _T_Orlov:
        leave_type = leave_info["leave_type"]
        return model(
            uuid=str(uuid),
            bvn=leave_info["user_key"],
            bruger_uuid=leave_info["user"],
            orlovstype_uuid=leave_type,
            orlovstype_titel=self.lc.classes[leave_type]["title"],
            engagement_uuid=leave_info["engagement"],
            startdato=leave_info["from_date"],
            slutdato=leave_info["to_date"],
        )

    def _add_leaves(self) -> None:
        logger.info("Add leaves")
        leaves = tqdm(self.lc.leaves.items(), desc="Export leave", unit="leave")
        for chunk in ichunked(leaves, self.chunk_size):
            for uuid, leave_validity in chunk:
                for leave_info in leave_validity:
                    sql_leave = self._generate_sql_leave(uuid, leave_info, WOrlov)
                    self.session.add(sql_leave)
            self.session.commit()

    def _generate_sql_it_systems(
        self, uuid, itsystem_info, model: Type[_T_ItSystem]
    ) -> _T_ItSystem:
        return model(uuid=str(uuid), navn=itsystem_info["name"])

    def _add_it_systems(self) -> None:
        logger.info("Add IT systems")
        itsystems = tqdm(
            self.lc.itsystems.items(), desc="Export itsystem", unit="itsystem"
        )
        for chunk in ichunked(itsystems, self.chunk_size):
            for uuid, itsystem_info in chunk:
                sql_itsystem = self._generate_sql_it_systems(
                    uuid, itsystem_info, WItSystem
                )
                self.session.add(sql_itsystem)
            self.session.commit()

    def _generate_sql_it_user(
        self, uuid, it_connection_info, model: Type[_T_ItForbindelse]
    ) -> _T_ItForbindelse:
        return model(
            uuid=str(uuid),
            it_system_uuid=it_connection_info["itsystem"],
            bruger_uuid=it_connection_info["user"],
            enhed_uuid=it_connection_info["unit"],
            brugernavn=it_connection_info["username"],
            startdato=it_connection_info["from_date"],
            slutdato=it_connection_info["to_date"],
            primær_boolean=it_connection_info.get("primary_boolean"),
        )

    def _add_it_users(self):
        logger.info("Add IT users")
        it_connections = tqdm(
            self.lc.it_connections.items(),
            desc="Export it connection",
            unit="it connection",
        )
        for chunk in ichunked(it_connections, self.chunk_size):
            for uuid, it_connection_validity in chunk:
                for it_connection_info in it_connection_validity:
                    sql_it_connection = self._generate_sql_it_user(
                        uuid, it_connection_info, WItForbindelse
                    )
                    self.session.add(sql_it_connection)
            self.session.commit()

    def _generate_sql_kle(self, uuid, kle_info, model: Type[_T_KLE]) -> _T_KLE:
        return model(
            uuid=str(uuid),
            enhed_uuid=kle_info["unit"],
            kle_aspekt_uuid=kle_info["kle_aspect"],
            kle_aspekt_titel=self.lc.classes[kle_info["kle_aspect"]]["title"],
            kle_nummer_uuid=kle_info["kle_number"],
            kle_nummer_titel=self.lc.classes[kle_info["kle_number"]]["title"],
            startdato=kle_info["from_date"],
            slutdato=kle_info["to_date"],
        )

    def _add_kles(self) -> None:
        logger.info("Add KLES")
        kles = tqdm(self.lc.kles.items(), desc="Export KLE", unit="KLE")
        for chunk in ichunked(kles, self.chunk_size):
            for uuid, kle_validity in chunk:
                for kle_info in kle_validity:
                    sql_kle = self._generate_sql_kle(uuid, kle_info, WKLE)
                    self.session.add(sql_kle)
            self.session.commit()

    def _add_receipt(self, query_time, start_time=None, end_time=None):
        logger.info("Add Receipt")
        sql_kvittering = Kvittering(
            query_tid=query_time,
            start_levering_tid=start_time,
            slut_levering_tid=end_time,
        )
        self.session.add(sql_kvittering)
        self.session.commit()
        return sql_kvittering

    def _update_receipt(self, sql_kvittering, start_time=None, end_time=None):
        logger.info("Update Receipt")
        sql_kvittering.start_levering_tid = start_time
        sql_kvittering.slut_levering_tid = end_time
        self.session.commit()

    def _generate_sql_related(
        self, uuid, related_info, model: Type[_T_Enhedssammenkobling]
    ) -> _T_Enhedssammenkobling:
        return model(
            uuid=str(uuid),
            enhed1_uuid=related_info["unit1_uuid"],
            enhed2_uuid=related_info["unit2_uuid"],
            startdato=related_info["from_date"],
            slutdato=related_info["to_date"],
        )

    def _add_related(self) -> None:
        logger.info("Add Enhedssammenkobling")
        relateds = tqdm(self.lc.related.items(), desc="Export related", unit="related")
        for chunk in ichunked(relateds, self.chunk_size):
            for uuid, related_validity in chunk:
                for related_info in related_validity:
                    sql_related = self._generate_sql_related(
                        uuid, related_info, WEnhedssammenkobling
                    )
                    self.session.add(sql_related)
            self.session.commit()

    def _generate_sql_managers(
        self, uuid, manager_info, model: Type[_T_Leder]
    ) -> _T_Leder:
        return model(
            uuid=str(uuid),
            bruger_uuid=manager_info["user"],
            enhed_uuid=manager_info["unit"],
            niveautype_uuid=manager_info["manager_level"],
            ledertype_uuid=manager_info["manager_type"],
            niveautype_titel=self.lc.classes[manager_info["manager_level"]]["title"],
            ledertype_titel=self.lc.classes[manager_info["manager_type"]]["title"],
            startdato=manager_info["from_date"],
            slutdato=manager_info["to_date"],
        )

    def _generate_sql_manager_responsibility(
        self, uuid, manager_uuid, manager_info, model: Type[_T_LederAnsvar]
    ) -> _T_LederAnsvar:
        return model(
            leder_uuid=str(manager_uuid),
            lederansvar_uuid=uuid,
            lederansvar_titel=self.lc.classes[uuid]["title"],
            startdato=manager_info["from_date"],
            slutdato=manager_info["to_date"],
        )

    def _add_managers(self) -> None:
        logger.info("Add managers")
        managers = tqdm(self.lc.managers.items(), desc="Export manager", unit="manager")
        for chunk in ichunked(managers, self.chunk_size):
            for manager_uuid, manager_validity in chunk:
                for manager_info in manager_validity:
                    sql_manager = self._generate_sql_managers(
                        manager_uuid, manager_info, WLeder
                    )
                    self.session.add(sql_manager)

                    for responsibility_uuid in manager_info["manager_responsibility"]:
                        sql_responsibility = self._generate_sql_manager_responsibility(
                            responsibility_uuid,
                            manager_uuid,
                            manager_info,
                            WLederAnsvar,
                        )
                        self.session.add(sql_responsibility)
            self.session.commit()

    def log_overlapping_runs_aak(self):
        self.engine.execute(
            "INSERT INTO [dbo].[kvittering_afvigelse] "
            "([query_tid],[aarsag]) VALUES (getdate(), "
            "'Time-export: hopper over da foregående "
            "loop stadig kører.')"
        )

    def export(self, resolve_dar: bool, use_pickle: typing.Any) -> None:
        self.perform_export(
            resolve_dar=resolve_dar,
            use_pickle=use_pickle,
        )

        self.swap_tables()

    def update_sql(self, uuid: UUID, objects: list[sql_type], table: Type[sql_type]):
        """Updates sql with the provided objects matching the objects UUID.

        Given a UUID, a list of objects and a table  we find any objects currently in sql for the given uuid.
        Then we add the objects that are not allready in sql - either new or changed in MO -  and remove any that
        do not match.
        """
        search_key = table.leder_uuid if table == LederAnsvar else table.uuid
        # Lookup engagement in sql
        current_objects = self.session.execute(
            select(table).where(search_key == str(uuid))
        ).all()
        if current_objects is not None:
            current_objects = [one(c) for c in current_objects]

        unchanged = [n for n in current_objects if n in objects]

        # Delete all rows from sql that do not match the found objects
        removed = [r for r in current_objects if r not in objects]
        logger.info(f"Delete {len(removed)} rows to {table} for {uuid=}")
        for r in removed:
            self.session.delete(r)

        # Create all rows not currently in sql
        new = [n for n in objects if n not in current_objects]
        logger.info(f"Add {len(new)} rows to {table} for {uuid=}")
        for n in new:
            self.session.add(n)

        # Check that the result is the expected amount of rows in sql.
        assert len(objects) == len(unchanged) + len(
            new
        ), f"expected {len(objects)=} to be equal to {len(unchanged)=} + {len(new)=}"

        self.session.commit()


def wrap_export(args: dict, settings: dict) -> None:
    sql_export = SqlExport(
        force_sqlite=args["force_sqlite"],
        historic=args["historic"],
        settings=settings,
    )
    try:
        lock_name = "sql_export_actual"

        if args["historic"]:
            lock_name = "sql_export_historic"

        ensure_single_run(
            func=sql_export.export,
            lock_name=lock_name,
            resolve_dar=args["resolve_dar"],
            use_pickle=args["read_from_cache"],
        )

    except ra_utils.ensure_single_run.LockTaken as name_of_lock:
        logger.warning(f"Lock {name_of_lock} taken, aborting export")
        if "log_overlapping_aak" in settings and settings.get("log_overlapping_aak"):
            sql_export.log_overlapping_runs_aak()


@click.command(help="SQL export")
@click.option("--resolve-dar", is_flag=True, envvar="RESOLVE_DAR")
@click.option("--historic", is_flag=True)
@click.option("--read-from-cache", is_flag=True, envvar="USE_CACHED_LORACACHE")
@click.option("--force-sqlite", is_flag=True)
def cli(**args):
    """
    Command line interface.
    """
    pydantic_settings = SqlExportSettings()
    pydantic_settings.start_logging_based_on_settings()
    logger.info("Command line args: %r", args)

    settings = load_settings()

    wrap_export(args=args, settings=settings)

    logger.info("*SQL export ended*")


if __name__ == "__main__":
    cli()



import logging
import sys
from datetime import datetime
from operator import itemgetter
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional

import requests
from more_itertools import only
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_settings
from ra_utils.tqdm_wrapper import tqdm
from requests import Session

import constants
from integrations import dawa_helper
from integrations.ad_integration import ad_reader
from integrations.opus import opus_helpers
from integrations.opus import payloads
from integrations.opus.opus_exceptions import EmploymentIdentifierNotUnique
from integrations.opus.opus_exceptions import RunDBInitException
from integrations.opus.opus_exceptions import UnknownOpusUnit

logger = logging.getLogger("opusDiff")

LOG_LEVEL = logging.DEBUG

logger = logging.getLogger("opusImport")

for name in logging.root.manager.loggerDict:
    if name in (
        "opusImport",
        "opusHelper",
        "opusDiff",
        "moImporterMoraTypes",
        "moImporterMoxTypes",
        "moImporterUtilities",
        "moImporterHelpers",
        "ADReader",
    ):
        logging.getLogger(name).setLevel(LOG_LEVEL)
    else:
        logging.getLogger(name).setLevel(logging.ERROR)

logging.basicConfig(
    format="%(levelname)s %(asctime)s %(name)s %(message)s",
    level=LOG_LEVEL,
    stream=sys.stdout,
)
UNIT_ADDRESS_CHECKS = {
    "seNr": constants.addresses_unit_se,
    "cvrNr": constants.addresses_unit_cvr,
    "eanNr": constants.addresses_unit_ean,
    "pNr": constants.addresses_unit_pnr,
    "phoneNumber": constants.addresses_unit_phoneNumber,
    "dar": constants.addresses_unit_dar,
}

EMPLOYEE_ADDRESS_CHECKS = {
    "phone": constants.addresses_employee_phone,
    "email": constants.addresses_employee_email,
    "dar": constants.addresses_employee_dar,
}
predefined_scopes = {
    constants.addresses_employee_dar: "DAR",
    constants.addresses_employee_phone: "PHONE",
    constants.addresses_employee_email: "EMAIL",
    constants.addresses_unit_dar: "DAR",
    constants.addresses_unit_phoneNumber: "PHONE",
    constants.addresses_unit_ean: "EAN",
    constants.addresses_unit_cvr: "TEXT",
    constants.addresses_unit_pnr: "TEXT",
    constants.addresses_unit_se: "TEXT",
}


class OpusDiffImport(object):
    def __init__(self, xml_date, ad_reader, employee_mapping={}, filter_ids={}):
        logger.info("Opus diff importer __init__ started")
        self.xml_date = xml_date
        self.ad_reader = ad_reader
        self.employee_forced_uuids = employee_mapping or opus_helpers.read_cpr_mapping()

        self.settings = load_settings()
        self.filter_ids = filter_ids or self.settings.get(
            "integrations.opus.units.filter_ids", []
        )

        self.session = Session()
        self.helper = self._get_mora_helper(
            hostname=self.settings["mora.base"], use_cache=False
        )
        try:
            self.org_uuid = self.helper.read_organisation()
        except KeyError:
            msg = "No root organisation in MO"
            logger.warning(msg)
            print(msg)
            return
        except requests.exceptions.RequestException as e:
            logger.error(e)
            print(e)
            exit()

        it_systems = self.helper.read_it_systems()
        self.it_systems = dict(map(itemgetter("name", "uuid"), it_systems))

        logger.info("__init__ done, now ready for import")

    def ensure_class_in_facet(self, *args, **kwargs):
        """Helper function to call ensure_class_in_facet from morahelpers with owner"""
        return self.helper.ensure_class_in_facet(
            *args, owner=opus_helpers.find_opus_root_unit_uuid(), **kwargs
        )

    def _find_classes(self, facet):
        class_types = self.helper.read_classes_in_facet(facet)
        types_dict = {}
        facet = class_types[1]
        for class_type in class_types[0]:
            types_dict[class_type["user_key"]] = class_type["uuid"]
        return types_dict, facet

    def _get_mora_helper(self, hostname="localhost:5000", use_cache=False):
        return MoraHelper(hostname=self.settings["mora.base"], use_cache=False)

    # This exact function also exists in sd_changed_at
    def _assert(self, response):
        """Check response is as expected"""
        assert response.status_code in (200, 400, 404)
        if response.status_code == 400:
            # Check actual response
            assert (
                "not give raise to a new registration" in response.text
            ), response.text
            logger.info("Requst had no effect")
        return None

    def _get_organisationfunktion(self, lora_uuid):
        resource = "/organisation/organisationfunktion/{}"
        resource = resource.format(lora_uuid)
        response = self.session.get(url=self.settings["mox.base"] + resource)
        response.raise_for_status()
        data = response.json()
        data = data[lora_uuid][0]["registreringer"][0]
        # logger.debug('Organisationsfunktionsinfo: {}'.format(data))
        return data

    def _find_engagement(self, bvn, funktionsnavn, present=False):
        resource = "/organisation/organisationfunktion?bvn={}&funktionsnavn={}".format(
            bvn, funktionsnavn
        )
        if present:
            resource += "&gyldighed=Aktiv"
        response = self.session.get(url=self.settings["mox.base"] + resource)
        response.raise_for_status()
        uuids = response.json()["results"][0]
        if len(uuids) > 1:
            msg = "Employment ID {} not unique: {}".format(bvn, uuids)
            logger.error(msg)
            raise EmploymentIdentifierNotUnique(msg)

        logger.info("bvn: {}, uuid: {}".format(bvn, uuids))
        if uuids:
            return uuids[0]

    def validity(self, employee, edit=False):
        """
        Calculates a validity object from en employee object.
        :param employee: An Opus employee object.
        :param edit: If True from will be current dump date, if true
        from will be taken from emploee object.
        :return: A valid MO valididty payload
        """
        to_date = employee["leaveDate"]
        # if to_date is None: # This can most likely be removed
        #     to_datetime = datetime.strptime('9999-12-31', '%Y-%m-%d')
        # else:
        #     to_datetime = datetime.strptime(to_date, '%Y-%m-%d')

        from_date = employee["entryDate"]
        if (from_date is None) or edit:
            from_date = self.xml_date.strftime("%Y-%m-%d")
        validity = {"from": from_date, "to": to_date}
        return validity

    def _condense_employee_mo_addresses(self, mo_uuid):
        """
        Read all addresses from MO an return as a simple dict
        """
        # Unfortunately, mora-helper currently does not read all addresses
        user_addresses = self.helper._mo_lookup(mo_uuid, "e/{}/details/address")
        address_dict = {}  # Condensate of all MO addresses for the employee
        if not isinstance(user_addresses, list):
            # In case the request to mo fails we assume no addresses in MO.
            # This has happend when the something in lora has been deleted.
            return address_dict

        for address in user_addresses:
            if address_dict.get(address["address_type"]["uuid"]) is not None:
                # More than one of this type exist in MO, this is not allowed.
                msg = "Inconsistent addresses for employee: {}"
                logger.error(msg.format(mo_uuid))
            address_dict[address["address_type"]["uuid"]] = {
                "value": address["value"],
                "uuid": address["uuid"],
            }
        return address_dict

    def _condense_employee_opus_addresses(self, employee):
        opus_addresses = {}
        if "email" in employee and not self.settings.get(
            "integrations.opus.skip_employee_email", False
        ):
            opus_addresses["email"] = employee["email"]

        skip_employee_phone = self.settings.get(
            "integrations.opus.skip_employee_phone", False
        )
        if employee["workPhone"] is not None:
            phone = opus_helpers.parse_phone(employee["workPhone"])
            if (phone is not None) and (not skip_employee_phone):
                opus_addresses["phone"] = phone

        if (
            "postalCode" in employee
            and employee["address"]
            and not self.settings.get("integrations.opus.skip_employee_address", False)
        ):
            if isinstance(employee["address"], dict):
                logger.info("Protected addres, cannont import")
            else:
                address_string = employee["address"]
                zip_code = employee["postalCode"]
                address_uuid = dawa_helper.dawa_lookup(address_string, zip_code)
                if address_uuid:
                    opus_addresses["dar"] = address_uuid
                else:
                    logger.warning("Could not find address in DAR")
        return opus_addresses

    def _perform_address_update(self, args, current):
        addr_type = args["address_type"]["uuid"]
        if current is None:  # Create address
            payload = payloads.create_address(**args)
            logger.debug("Create {} address payload: {}".format(addr_type, payload))
            response = self.helper._mo_post("details/create", payload)
            assert response.status_code == 201
        elif (args.get("value") == current.get("value")) and (
            args.get("visibility") == current.get("visibility")
        ):  # Nothing changed
            logger.info("{} not updated".format(addr_type))
        else:  # Edit address
            payload = payloads.edit_address(args, current["uuid"])
            logger.debug("Edit address {}, payload: {}".format(addr_type, payload))
            response = self.helper._mo_post("details/edit", payload)
            response.raise_for_status()

    def _update_employee_address(self, mo_uuid, employee):
        opus_addresses = self._condense_employee_opus_addresses(employee)
        mo_addresses = self._condense_employee_mo_addresses(mo_uuid)
        logger.info("Addresses to be synced to MO: {}".format(opus_addresses))

        for addr_type, mo_addr_type in EMPLOYEE_ADDRESS_CHECKS.items():
            if opus_addresses.get(addr_type) is None:
                continue

            addr_type_uuid = self.ensure_class_in_facet(
                "employee_address_type",
                mo_addr_type,
                scope=predefined_scopes.get(mo_addr_type),
            )
            visibility = None
            if mo_addr_type == "Adresse":
                visibility = self.ensure_class_in_facet(
                    facet="visibility",
                    bvn="Intern",
                    title="Må vises internt",
                    scope="INTERNAL",
                )

            current = mo_addresses.get(str(addr_type_uuid))
            address_args = {
                "address_type": {"uuid": str(addr_type_uuid)},
                "value": opus_addresses[addr_type],
                "validity": {"from": self.xml_date.strftime("%Y-%m-%d"), "to": None},
                "user_uuid": mo_uuid,
            }
            if visibility:
                address_args["visibility"] = {"uuid": str(visibility)}
            self._perform_address_update(address_args, current)

    def _update_unit_addresses(self, unit):
        calculated_uuid = opus_helpers.generate_uuid(unit["@id"])
        unit_addresses = self.helper.read_ou_address(
            calculated_uuid, scope=None, return_all=True
        )

        address_dict = {}
        for address in unit_addresses:
            if address_dict.get(address["type"]) is not None:
                # More than one of this type exist in MO, this is not allowed.
                msg = "Inconsistent addresses for unit: {}"
                logger.error(msg.format(calculated_uuid))
            # if address['value'] not in ('9999999999999', '0000000000'):
            address_dict[address["type"]] = {
                "value": address["value"],
                "uuid": address["uuid"],
            }

        if unit.get("street") and unit.get("zipCode"):
            address_uuid = dawa_helper.dawa_lookup(unit["street"], unit["zipCode"])
            if address_uuid:
                logger.debug("Found DAR uuid: {}".format(address_uuid))
                unit["dar"] = address_uuid
            else:
                logger.warning(
                    "Failed to lookup {}, {}".format(unit["street"], unit["zipCode"])
                )

        for addr_type, mo_addr_type in UNIT_ADDRESS_CHECKS.items():
            # addr_type is the opus name for the address, mo_addr_type
            # is read from MO
            if unit.get(addr_type) is None:
                continue

            addr_type_uuid = self.ensure_class_in_facet(
                "org_unit_address_type",
                mo_addr_type,
                scope=predefined_scopes.get(mo_addr_type),
            )
            current = address_dict.get(str(addr_type_uuid))
            args = {
                "address_type": {"uuid": str(addr_type_uuid)},
                "value": unit[addr_type],
                "validity": {"from": self.xml_date.strftime("%Y-%m-%d"), "to": None},
                "unit_uuid": str(calculated_uuid),
            }
            self._perform_address_update(args, current)

    def update_unit(self, unit):
        calculated_uuid = opus_helpers.generate_uuid(unit["@id"])
        parent_name = unit["parentOrgUnit"]
        parent_uuid = (
            opus_helpers.generate_uuid(parent_name)
            if parent_name
            else self.helper.read_organisation()
        )
        mo_unit = self.helper.read_ou(calculated_uuid)

        # Default to "Enhed" as unittype
        org_type_title = unit.get("orgTypeTxt", "Enhed")
        org_type_bvn = unit.get("orgType", org_type_title)

        unit_type = self.ensure_class_in_facet(
            "org_unit_type", bvn=org_type_bvn, title=org_type_title
        )
        from_date = unit.get("startDate", "01-01-1900")
        unit_user_key = self.settings.get("integrations.opus.unit_user_key", "@id")
        unit_args = {
            "unit": unit,
            "unit_user_key": unit[unit_user_key],
            "unit_uuid": str(calculated_uuid),
            "unit_type": str(unit_type),
            "parent": str(parent_uuid),
            "from_date": from_date,
        }

        if mo_unit.get("uuid"):  # Edit
            unit_args["from_date"] = self.xml_date.strftime("%Y-%m-%d")
            payload = payloads.edit_org_unit(**unit_args)
            logger.info("Edit unit: {}".format(payload))
            response = self.helper._mo_post("details/edit", payload)
            if response.status_code == 400:
                assert response.text.find("raise to a new registration") > 0
            else:
                response.raise_for_status()
        else:  # Create
            payload = payloads.create_org_unit(**unit_args)
            logger.debug("Create department payload: {}".format(payload))
            response = self.helper._mo_post("ou/create", payload)
            response.raise_for_status()
            logger.info("Created unit {}".format(unit["@id"]))
            logger.debug("Response: {}".format(response.text))

        self._update_unit_addresses(unit)

    def _job_and_engagement_type(self, employee):
        job = employee["position"]
        job_function_uuid = self.ensure_class_in_facet(
            "engagement_job_function", bvn=job
        )

        contract = employee.get("workContractText", "Ansat")
        engagement_type_uuid = self.ensure_class_in_facet("engagement_type", contract)
        return str(job_function_uuid), str(engagement_type_uuid)

    def update_engagement(self, engagement, employee):
        """
        Update a MO engagement according to opus employee object.
        It often happens that the change that provoked lastChanged to
        be updated is not a MO field, and thus we check for relevant
        differences before shipping the payload to MO.
        :param engagement: Relevant MO engagement object.
        :param employee: Relevent Opus employee object.
        :return: True if update happended, False if not.
        """
        job_function, eng_type = self._job_and_engagement_type(employee)
        unit_uuid = opus_helpers.generate_uuid(employee["orgUnit"])

        validity = self.validity(employee, edit=True)
        data = {
            "engagement_type": {"uuid": eng_type},
            "job_function": {"uuid": job_function},
            "org_unit": {"uuid": str(unit_uuid)},
            "validity": validity,
        }

        engagement_unit = self.helper.read_ou(unit_uuid)
        if "error" in engagement_unit:
            msg = "The wanted unit does not exit: {}"
            logger.error(msg.format(unit_uuid))
            raise UnknownOpusUnit

        if engagement["validity"]["to"] is None:
            old_valid_to = datetime.strptime("9999-12-31", "%Y-%m-%d")
        else:
            old_valid_to = datetime.strptime(engagement["validity"]["to"], "%Y-%m-%d")
        if validity["to"] is None:
            new_valid_to = datetime.strptime("9999-12-31", "%Y-%m-%d")
        else:
            new_valid_to = datetime.strptime(validity["to"], "%Y-%m-%d")

        something_new = not (
            (engagement["engagement_type"]["uuid"] == eng_type)
            and (engagement["job_function"]["uuid"] == job_function)
            and (engagement["org_unit"]["uuid"] == str(unit_uuid))
            and (old_valid_to == new_valid_to)
        )

        logger.info("Something new? {}".format(something_new))
        if something_new:
            payload = payloads.edit_engagement(data, engagement["uuid"])
            logger.debug("Update engagement payload: {}".format(payload))
            response = self.helper._mo_post("details/edit", payload)
            self._assert(response)

        if new_valid_to < old_valid_to:
            self.terminate_detail(
                engagement["uuid"], detail_type="engagement", end_date=new_valid_to
            )
        return something_new

    def create_engagement(self, mo_user_uuid, opus_employee):
        job_function, eng_type = self._job_and_engagement_type(opus_employee)
        unit_uuid = opus_helpers.generate_uuid(opus_employee["orgUnit"])

        engagement_unit = self.helper.read_ou(unit_uuid)
        if "error" in engagement_unit:
            msg = "The wanted unit does not exit: {}"
            logger.error(msg.format(opus_employee["orgUnit"]))
            raise UnknownOpusUnit

        validity = self.validity(opus_employee, edit=False)
        payload = payloads.create_engagement(
            employee=opus_employee,
            user_uuid=mo_user_uuid,
            unit_uuid=unit_uuid,
            job_function=job_function,
            engagement_type=eng_type,
            validity=validity,
        )
        logger.debug("Create engagement payload: {}".format(payload))
        response = self.helper._mo_post("details/create", payload)
        assert response.status_code == 201

    def create_user(self, employee, uuid=None):
        payload = payloads.create_user(employee, self.org_uuid, uuid)

        logger.info("Create user payload: {}".format(payload))
        r = self.helper._mo_post("e/create", payload)
        r.raise_for_status()
        return_uuid = r.json()
        logger.info(
            "Created employee {} {} with uuid {}".format(
                employee["firstName"], employee["lastName"], return_uuid
            )
        )

        return return_uuid

    def connect_it_system(self, username, it_system, employee, person_uuid):
        it_system_uuid = self.it_systems[it_system]
        current = self.helper.get_e_itsystems(
            person_uuid, it_system_uuid=it_system_uuid
        )
        try:
            current = only(current, default={})
        except ValueError:
            logger.warning(
                f"Skipped connecting {it_system} IT system . More than one IT system found for {person_uuid=}"
            )
            return

        if not (username or current):
            return
        # New it-system account
        if not current:
            payload = payloads.connect_it_system_to_user(
                username,
                it_system_uuid,
                person_uuid,
                self.xml_date.strftime("%Y-%m-%d"),
            )
            logger.debug(f"{it_system} account payload: {payload}")
            response = self.helper._mo_post("details/create", payload)
            assert response.status_code == 201
            logger.info(f"Added {it_system} info for {person_uuid}")
        # Deleted it-system account
        elif not username:
            self.terminate_detail(current["uuid"], detail_type="it")
            logger.info(f"No {it_system} info for {person_uuid} any longer")
        # Changed account name. Only supports one account pr it-system
        elif current.get("user_key") != username:
            payload = payloads.edit_it_system_username(
                current["uuid"],
                username,
                self.xml_date.strftime("%Y-%m-%d"),
            )
            logger.debug(f"{it_system} account payload: {payload}")
            response = self.helper._mo_post("details/edit", payload)
            response.raise_for_status()
            logger.info(f"Changed {it_system} info for {person_uuid}")

    def _to_datetime(self, item):
        if item is None:
            item_datetime = datetime.strptime("9999-12-31", "%Y-%m-%d")
        else:
            item_datetime = datetime.strptime(item, "%Y-%m-%d")
        return item_datetime

    def update_manager_status(self, employee_mo_uuid, employee):
        url = "e/{}/details/manager?at=" + self.validity(employee, edit=True)["from"]
        manager_functions = self.helper._mo_lookup(employee_mo_uuid, url)
        logger.debug("Manager functions to update: {}".format(manager_functions))
        if manager_functions:
            logger.debug("Manager functions to update: {}".format(manager_functions))

        if employee["isManager"] == "false":
            if manager_functions:
                logger.info("Terminate manager function")
                self.terminate_detail(
                    manager_functions[0]["uuid"], detail_type="manager"
                )
            else:
                logger.debug("Correctly not a manager")

        if employee["isManager"] == "true":
            manager_level = "{}.{}".format(
                employee["superiorLevel"], employee["subordinateLevel"]
            )
            manager_level_uuid = self.ensure_class_in_facet(
                "manager_level", manager_level
            )
            manager_type = employee["position"]
            manager_type_uuid = self.ensure_class_in_facet("manager_type", manager_type)
            responsibility_uuid = self.ensure_class_in_facet(
                "responsibility", "Lederansvar"
            )

            args = {
                "unit": str(opus_helpers.generate_uuid(employee["orgUnit"])),
                "person": employee_mo_uuid,
                "manager_type": str(manager_type_uuid),
                "level": str(manager_level_uuid),
                "responsibility": str(responsibility_uuid),
                "validity": self.validity(employee, edit=True),
            }
            if manager_functions:
                logger.info("Attempt manager update of {}:".format(employee_mo_uuid))
                # Currently Opus supports only a single manager object pr employee
                assert len(manager_functions) == 1

                mf = manager_functions[0]

                payload = payloads.edit_manager(
                    object_uuid=manager_functions[0]["uuid"], **args
                )

                something_new = not (
                    mf["org_unit"]["uuid"] == args["unit"]
                    and mf["person"]["uuid"] == args["person"]
                    and mf["manager_type"]["uuid"] == args["manager_type"]
                    and mf["manager_level"]["uuid"] == args["level"]
                    and mf["responsibility"][0]["uuid"] == args["responsibility"]
                )

                if something_new:
                    logger.debug("Something is changed, execute payload")
                else:
                    mo_end_datetime = self._to_datetime(mf["validity"]["to"])
                    opus_end_datetime = self._to_datetime(args["validity"]["to"])
                    logger.info("MO end datetime: {}".format(mo_end_datetime))
                    logger.info("OPUS end datetime: {}".format(opus_end_datetime))

                    if mo_end_datetime == opus_end_datetime:
                        logger.info("No edit of manager object")
                        payload = None
                    elif opus_end_datetime > mo_end_datetime:
                        logger.info("Extend validity, send payload to MO")
                    else:  # opus_end_datetime < mo_end_datetime:
                        logger.info("Terminate mangement role")
                        payload = None
                        self.terminate_detail(
                            mf["uuid"],
                            detail_type="manager",
                            end_date=opus_end_datetime,
                        )

                logger.debug("Update manager payload: {}".format(payload))
                if payload is not None:
                    response = self.helper._mo_post("details/edit", payload)
                    self._assert(response)
            else:  # No existing manager functions
                logger.info("Turn this person into a manager")
                # Validity is set to edit=True since the validiy should
                # calculated as an edit to the engagement
                payload = payloads.create_manager(user_key=employee["@id"], **args)
                logger.debug("Create manager payload: {}".format(payload))
                response = self.helper._mo_post("details/create", payload)
                assert response.status_code == 201

    def update_employee(self, employee):
        cpr = opus_helpers.read_cpr(employee)
        logger.info("----")
        logger.info("Now updating {}".format(employee.get("@id")))
        logger.debug("Available info: {}".format(employee))
        mo_user = self.helper.read_user(user_cpr=cpr, use_cache=False)

        ad_info = {}
        if self.ad_reader is not None:
            ad_info = self.ad_reader.read_user(cpr=cpr)

        if mo_user is None:
            uuid = self.employee_forced_uuids.get(cpr)
            logger.info(f"Employee {uuid} in force list. AD info: {ad_info}")
            if uuid is None:
                uuid = ad_info.get("ObjectGuid")
            employee_mo_uuid = self.create_user(employee, uuid)
        else:
            employee_mo_uuid = mo_user["uuid"]

            # Update user if name has changed
            if (employee["firstName"] != mo_user["givenname"]) or (
                employee["lastName"] != mo_user["surname"]
            ):
                employee_mo_uuid = self.create_user(employee, employee_mo_uuid)
                logger.info(f"Updated name of employee {employee_mo_uuid}")

        # Add it-systems
        self.connect_it_system(
            employee.get("userId"), constants.Opus_it_system, employee, employee_mo_uuid
        )

        if self.ad_reader is not None:
            sam_account = ad_info.get("SamAccountName")
            self.connect_it_system(
                sam_account, constants.AD_it_system, employee, employee_mo_uuid
            )

        self._update_employee_address(employee_mo_uuid, employee)

        # Now we have a MO uuid, update engagement:
        mo_engagements = self.helper.read_user_engagement(
            employee_mo_uuid, read_all=True
        )
        user_engagements = filter(
            lambda eng: eng["user_key"] == employee["@id"], mo_engagements
        )
        current_mo_eng = None
        for eng in user_engagements:
            current_mo_eng = eng["uuid"]
            val_from = datetime.strptime(eng["validity"]["from"], "%Y-%m-%d")
            val_to = datetime.strptime("9999-12-31", "%Y-%m-%d")
            if eng["validity"]["to"] is not None:
                val_to = datetime.strptime(eng["validity"]["to"], "%Y-%m-%d")
            if val_from < self.xml_date < val_to:
                logger.info("Found current validty {}".format(eng["validity"]))
                break

        if current_mo_eng is None:
            self.create_engagement(employee_mo_uuid, employee)
        else:
            logger.info("Validity for {}: {}".format(employee["@id"], eng["validity"]))
            self.update_engagement(eng, employee)

        self.update_manager_status(employee_mo_uuid, employee)

    def terminate_detail(self, uuid, detail_type="engagement", end_date=None):
        if end_date is None:
            end_date = self.xml_date

        payload = payloads.terminate_detail(
            uuid, end_date.strftime("%Y-%m-%d"), detail_type
        )
        logger.debug("Terminate payload: {}".format(payload))
        response = self.helper._mo_post("details/terminate", payload)
        logger.debug("Terminate response: {}".format(response.text))
        self._assert(response)

    def import_single_employment(self, employee):
        # logger.info('Update  employment {} from {}'.format(employment, xml_file))
        last_changed_str = employee.get("@lastChanged")
        if last_changed_str is not None:  # This is a true employee-object.
            self.update_employee(employee)
        else:  # This is an implicit termination.
            # This is a terminated employee, check if engagement is active
            # terminate if it is.
            if not employee["@action"] == "leave":
                msg = "Missing date on a non-leave object!"
                logger.error(msg)
                raise Exception(msg)

            org_funk_info = self._find_engagement(employee["@id"], present=True)
            if org_funk_info:
                logger.info("Terminating: {}".format(org_funk_info))
                self.terminate_detail(org_funk_info["engagement"])
                if "manager" in org_funk_info:
                    self.terminate_detail(
                        org_funk_info["manager"], detail_type="manager"
                    )

    def find_unterminated_filtered_units(self, units):
        """Check if units are in MO."""

        # Read all active MO org_units
        mo_units = self.helper.read_ou_root(
            org=self.helper.read_organisation(),
        )
        current_uuids = set(map(itemgetter("uuid"), mo_units))
        # return the units that are active in os2mo, but should be terminated
        mo_units = filter(
            lambda unit: opus_helpers.gen_unit_uuid(unit) in current_uuids, units
        )
        return mo_units

    def handle_filtered_units(self, units, dry_run=False):
        """Rules for handling units that are not imported from opus.

        If a unit is filtered from the Opus file it means it cannot be deleted in Opus, but should not appear in MO.
        Any units that exists in MO, but are later moved in Opus to be below one of the filtered units should be terminated in MO.
        """
        unfiltered_units = {opus_helpers.gen_unit_uuid(unit) for unit in units}
        if dry_run:
            print(
                f"There are {len(unfiltered_units)} units that should have been terminated."
            )
            if units:
                print(unfiltered_units)
            return

        for uuid in tqdm(unfiltered_units, desc="Terminating filtered org_units"):
            # TODO: Use actual org_unit terminate endpoint for this.
            self.terminate_detail(uuid, detail_type="org_unit", end_date=self.xml_date)

    def start_import(self, units, employees, terminated_employees):
        """
        Start an opus import, run the oldest available dump that
        has not already been imported.
        """

        for unit in tqdm(units, desc="Update units"):
            self.update_unit(unit)

        for employee in tqdm(employees, desc="Update employees"):
            self.update_employee(employee)

        for employee in tqdm(terminated_employees, desc="Terminating employees"):
            # This is a terminated employee, check if engagement is active
            # terminate if it is.
            if not employee["@action"] == "leave":
                msg = "This should be a terminated employee!"
                logger.error(msg)
                raise Exception(msg)

            eng_info = self._find_engagement(
                employee["@id"], "Engagement", present=True
            )
            if eng_info:
                logger.info("Terminating: {}".format(eng_info))
                self.terminate_detail(eng_info)
                manager_info = self._find_engagement(
                    employee["@id"], "Leder", present=True
                )
                if manager_info:
                    self.terminate_detail(manager_info, detail_type="manager")

        logger.info("Program ended correctly")


def import_one(
    ad_reader,
    xml_date: datetime,
    latest_date: Optional[datetime],
    dumps: Dict,
    filter_ids: List[str],
    opus_id: Optional[int] = None,
    rundb_write=True,
):
    """Import one file at the date xml_date."""
    msg = "Start update: File: {}, update since: {}"
    logger.info(msg.format(xml_date, latest_date))
    print(msg.format(xml_date, latest_date))
    # Find changes to units and employees
    latest_path = None
    if latest_date:
        latest_path = dumps[latest_date]
    xml_path = dumps[xml_date]
    (
        units,
        filtered_units,
        employees,
        terminated_employees,
    ) = opus_helpers.read_and_transform_data(
        latest_path, xml_path, filter_ids, opus_id=opus_id
    )
    if rundb_write:
        opus_helpers.local_db_insert((xml_date, "Running diff update since {}"))

    diff = OpusDiffImport(
        xml_date,
        ad_reader=ad_reader,
        filter_ids=filter_ids,
    )
    diff.start_import(units, employees, terminated_employees)
    filtered_units = diff.find_unterminated_filtered_units(filtered_units)

    diff.handle_filtered_units(filtered_units)
    if rundb_write:
        opus_helpers.local_db_insert((xml_date, "Diff update ended: {}"))
    print()


def start_opus_diff(ad_reader=None):
    """
    Start an opus update, use the oldest available dump that has not
    already been imported.
    """
    SETTINGS = load_settings()

    dumps = opus_helpers.read_available_dumps()
    run_db = Path(SETTINGS["integrations.opus.import.run_db"])
    filter_ids = SETTINGS.get("integrations.opus.units.filter_ids", [])

    if not run_db.is_file():
        logger.error("Local base not correctly initialized")
        raise RunDBInitException("Local base not correctly initialized")
    xml_date, latest_date = opus_helpers.next_xml_file(run_db, dumps)

    while xml_date:
        import_one(ad_reader, xml_date, latest_date, dumps, filter_ids, opus_id=None)
        # Check if there are more files to import
        xml_date, latest_date = opus_helpers.next_xml_file(run_db, dumps)
        logger.info("Ended update")


if __name__ == "__main__":
    settings = load_settings()

    reader = ad_reader.ADParameterReader() if settings.get("integrations.ad") else None

    try:
        start_opus_diff(ad_reader=reader)
    except RunDBInitException:
        print("RunDB not initialized")



from contextlib import contextmanager
from logging import ERROR
from unittest import mock
from unittest import TestCase

import pytest
from parameterized import parameterized

from .. import ad_common
from .. import sync_mo_uuid_to_ad
from ..ad_exceptions import ImproperlyConfigured
from .mocks import AD_UUID_FIELD
from .mocks import MO_UUID
from .mocks import MockADParameterReader
from .mocks import MockMoraHelper
from .mocks import MockUnknownCPRADParameterReader
from .mocks import UNKNOWN_CPR_NO


class _SyncMoUuidToAd(sync_mo_uuid_to_ad.SyncMoUuidToAd):
    """Testable subclass of the real `SyncMoUuidToAd` class."""

    def __init__(self, ad_cpr_no):
        self._ad_cpr_no = ad_cpr_no
        self._scripts = []
        super().__init__()

    def _get_mora_helper(self):
        return MockMoraHelper(self._ad_cpr_no)

    def _create_session(self):
        return mock.MagicMock()

    def _build_user_credential(self):
        return ""

    def _run_ps_script(self, ps_script):
        self._scripts.append(ps_script)


class _SyncMoUuidToAdRunPSScriptFails(_SyncMoUuidToAd):
    def _run_ps_script(self, ps_script):
        super()._run_ps_script(ps_script)
        raise Exception("an exception!")


# Based on this example:
# https://docs.pytest.org/en/stable/example/parametrize.html#parametrizing-conditional-raising


@contextmanager
def does_not_raise():
    yield


@pytest.mark.parametrize(
    "example_input,expectation",
    [
        ([{"properties": ["foo"]}], does_not_raise()),
        ([{"properties": ["bar", "baz"]}], pytest.raises(ImproperlyConfigured)),
        (
            [{"properties": ["bar", "baz"]}, {"properties": ["bar", "baz"]}],
            pytest.raises(ImproperlyConfigured),
        ),
        (
            [{"properties": ["foo", "baz"]}, {"properties": ["bar", "baz"]}],
            does_not_raise(),
        ),
    ],
)
def test_invalid_configuration(example_input, expectation):
    with expectation:
        TestSyncMoUuidToAd()._get_instance(
            {
                "integrations.ad.write.uuid_field": "foo",
                "integrations.ad": example_input,
            }
        )


class TestSyncMoUuidToAd(TestCase):
    """Test `sync_mo_uuid_to_ad`"""

    _ad_cpr_field_name = "extensionAttribute1"
    _mo_cpr_no = "0123456789"

    def test_sync_one(self):
        instance = self._get_instance()
        instance.sync_one(instance._ad_cpr_no)
        self._assert_script_contents_ok(instance)

    @parameterized.expand(
        [
            # Look up bogus CPR in mock AD which will not find an AD user
            (MockUnknownCPRADParameterReader(), "AD User not found"),
            # Look up bogus CPR in mock MO which will not find a MO user
            (MockADParameterReader(), "MO User not found"),
        ]
    )
    def test_sync_one_unknown_cpr_raises_exception(self, reader, expected_message):
        instance = self._get_instance(reader=reader)
        with self.assertRaisesRegex(Exception, expected_message):
            instance.sync_one(UNKNOWN_CPR_NO)

    def test_sync_one_uses_morahelper(self):
        instance = self._get_instance()
        instance.sync_one(self._mo_cpr_no)
        # Assert that our mocked `MoraHelper` recorded one call to `read_user`
        # with the expected CPR as its only argument.
        self.assertListEqual(instance.helper._read_user_calls, [self._mo_cpr_no])

    def test_sync_all(self):
        instance = self._get_instance()
        instance.sync_all()
        self._assert_script_contents_ok(instance)

    def test_perform_sync_handles_run_ps_script_exception(self):
        ad_users = [
            {
                self._ad_cpr_field_name: self._mo_cpr_no,
                "SamAccountName": "example",
            }
        ]
        mo_users = {self._mo_cpr_no: MO_UUID}
        instance = self._get_instance(cls=_SyncMoUuidToAdRunPSScriptFails)
        with self.assertLogs("MoUuidAdSync", ERROR) as cm:
            instance.perform_sync(ad_users, mo_users)
            self.assertIn("failed to write MO UUID", cm.records[0].message)

    def _get_instance(self, settings=None, reader=None, cls=_SyncMoUuidToAd):
        _settings = {
            "integrations.ad.write.uuid_field": AD_UUID_FIELD,
            "integrations.ad": [{"properties": [AD_UUID_FIELD]}],
            "primary": {"cpr_field": self._ad_cpr_field_name, "cpr_separator": "-"},
            "global": {},
        }

        if settings:
            _settings.update(settings)

        read_settings_mock = mock.patch.object(
            ad_common,
            "read_settings",
            return_value=_settings,
        )

        load_settings_mock = mock.patch.object(
            sync_mo_uuid_to_ad,
            "load_settings",
            return_value=_settings,
        )

        reader = reader or MockADParameterReader()

        reader_mock = mock.patch.object(
            sync_mo_uuid_to_ad,
            "ADParameterReader",
            new=lambda: reader,
        )

        ad_cpr_no = reader.read_user()[self._ad_cpr_field_name]

        with read_settings_mock:
            with load_settings_mock:
                with reader_mock:
                    instance = cls(ad_cpr_no)
                    return instance

    def _assert_script_contents_ok(self, instance):
        # Assert we tried to execute exactly one script, with the proper
        # contents.
        self.assertEqual(len(instance._scripts), 1)
        self.assertIn(
            '-Replace @{"%s"="%s"}' % (AD_UUID_FIELD, MO_UUID),
            instance._scripts[0],
        )
